#
# This is an extension to the Nautilus file manager to allow better
# integration with the Subversion source control system.
#
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2008 by Adam Plumb <adamplumb@gmail.com>
#
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

"""
Concrete VCS implementation for Subversion functionality.
"""
import subprocess
import os
import shutil
import os.path
from os.path import isdir, isfile, dirname, islink, realpath
from datetime import datetime

import pysvn

import rabbitvcs.vcs
import rabbitvcs.vcs.status
import rabbitvcs.vcs.log
import rabbitvcs.util.helper
from rabbitvcs.util.log import Log

log = Log("rabbitvcs.vcs.svn")

from rabbitvcs import gettext
_ = gettext.gettext

# Extra "action" for "commit completed"
commit_completed = "commit_completed"

class Revision:
    """
    Implements a simple revision object as a wrapper around the pysvn revision
    object.  This allows us to provide a standard interface to the object data.
    """

    KINDS = {
        "unspecified":      pysvn.opt_revision_kind.unspecified,
        "number":           pysvn.opt_revision_kind.number,
        "date":             pysvn.opt_revision_kind.date,
        "committed":        pysvn.opt_revision_kind.committed,
        "previous":         pysvn.opt_revision_kind.previous,
        "working":          pysvn.opt_revision_kind.working,
        "head":             pysvn.opt_revision_kind.head,
        "base":             pysvn.opt_revision_kind.base
    }

    def __init__(self, kind, value=None):
        self.kind = unicode(kind).lower()
        self.value = value
        self.is_revision_object = True

        # TODO Shift to options factory?
        if self.value and str(self.value).lower().strip() == 'head':
            self.kind = "head"

        if self.value is None and self.kind in ("number", "date"):
            self.kind = "head"

        if self.kind == "head":
            self.value = None

        self.__revision_kind = self.KINDS[self.kind]
        self.__revision = None

        try:
            if self.value is not None:
                self.__revision = pysvn.Revision(self.__revision_kind, self.value)
            else:
                self.__revision = pysvn.Revision(self.__revision_kind)
        except Exception, e:
            log.exception(e)

    def __unicode__(self):
        if self.value:
            return unicode(self.value)
        else:
            return self.kind

    def short(self):
        return self.__unicode__()

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return self.__unicode__()

    def primitive(self):
        return self.__revision

class SVN:
    """

    """

    STATUS = {
        "none"          : pysvn.wc_status_kind.none,
        "unversioned"   : pysvn.wc_status_kind.unversioned,
        "normal"        : pysvn.wc_status_kind.normal,
        "added"         : pysvn.wc_status_kind.added,
        "missing"       : pysvn.wc_status_kind.missing,
        "deleted"       : pysvn.wc_status_kind.deleted,
        "replaced"      : pysvn.wc_status_kind.replaced,
        "modified"      : pysvn.wc_status_kind.modified,
        "merged"        : pysvn.wc_status_kind.merged,
        "conflicted"    : pysvn.wc_status_kind.conflicted,
        "ignored"       : pysvn.wc_status_kind.ignored,
        "obstructed"    : pysvn.wc_status_kind.obstructed,
        "external"      : pysvn.wc_status_kind.external,
        "incomplete"    : pysvn.wc_status_kind.incomplete
    }

    STATUSES_FOR_COMMIT = map(str,
        [
            pysvn.wc_status_kind.unversioned,
            pysvn.wc_status_kind.added,
            pysvn.wc_status_kind.deleted,
            pysvn.wc_status_kind.replaced,
            pysvn.wc_status_kind.modified,
            pysvn.wc_status_kind.missing,
            pysvn.wc_status_kind.obstructed
        ])

    STATUSES_FOR_REVERT = map(str,
        [
            pysvn.wc_status_kind.missing,
            pysvn.wc_status_kind.added,
            pysvn.wc_status_kind.modified,
            pysvn.wc_status_kind.deleted
        ])

    STATUSES_FOR_ADD = map(str,
        [
            pysvn.wc_status_kind.unversioned,
            pysvn.wc_status_kind.obstructed
        ])

    STATUSES_FOR_RESOLVE = map(str,
        [
            pysvn.wc_status_kind.conflicted
        ])

    STATUSES_FOR_CHECK = map(str,
        [
            pysvn.wc_status_kind.added,
            pysvn.wc_status_kind.deleted,
            pysvn.wc_status_kind.replaced,
            pysvn.wc_status_kind.modified,
            pysvn.wc_status_kind.missing,
            pysvn.wc_status_kind.conflicted,
        ])

    PROPERTIES = {
        "executable":   "svn:executable",
        "mime-type":    "svn:mime-type",
        "ignore":       "svn:ignore",
        "keywords":     "svn:keywords",
        "eol-style":    "svn:eol-style",
        "externals":    "svn:externals",
        "special":      "svn:special"
    }

    NOTIFY_ACTIONS = {
        pysvn.wc_notify_action.add:                     _("Added"),
        pysvn.wc_notify_action.copy:                    _("Copied"),
        pysvn.wc_notify_action.delete:                  _("Deleted"),
        pysvn.wc_notify_action.restore:                 _("Restored"),
        pysvn.wc_notify_action.revert:                  _("Reverted"),
        pysvn.wc_notify_action.failed_revert:           _("Failed Revert"),
        pysvn.wc_notify_action.resolved:                _("Resolved"),
        pysvn.wc_notify_action.skip:                    _("Skipped"),
        pysvn.wc_notify_action.update_delete:           _("Deleted"),
        pysvn.wc_notify_action.update_add:              _("Added"),
        pysvn.wc_notify_action.update_update:           _("Updated"),
        pysvn.wc_notify_action.update_completed:        _("Completed"),
        pysvn.wc_notify_action.update_external:         _("External"),
        pysvn.wc_notify_action.status_completed:        _("Completed"),
        pysvn.wc_notify_action.status_external:         _("External"),
        pysvn.wc_notify_action.commit_modified:         _("Modified"),
        pysvn.wc_notify_action.commit_added:            _("Added"),
        pysvn.wc_notify_action.commit_deleted:          _("Deleted"),
        pysvn.wc_notify_action.commit_replaced:         _("Replaced"),
        pysvn.wc_notify_action.commit_postfix_txdelta:  _("Changed"),
        pysvn.wc_notify_action.annotate_revision:       _("Annotated"),
        pysvn.wc_notify_action.locked:                  _("Locked"),
        pysvn.wc_notify_action.unlocked:                _("Unlocked"),
        pysvn.wc_notify_action.failed_lock:             _("Failed Lock"),
        pysvn.wc_notify_action.failed_unlock:           _("Failed Unlock")
    }

    NOTIFY_ACTIONS_COMPLETE = [
        pysvn.wc_notify_action.status_completed,
        pysvn.wc_notify_action.update_completed,
        commit_completed
    ]

    NOTIFY_STATES = {
        pysvn.wc_notify_state.inapplicable:             _("Inapplicable"),
        pysvn.wc_notify_state.unknown:                  _("Unknown"),
        pysvn.wc_notify_state.unchanged:                _("Unchanged"),
        pysvn.wc_notify_state.missing:                  _("Missing"),
        pysvn.wc_notify_state.obstructed:               _("Obstructed"),
        pysvn.wc_notify_state.changed:                  _("Changed"),
        pysvn.wc_notify_state.merged:                   _("Merged"),
        pysvn.wc_notify_state.conflicted:               _("Conflicted")
    }

    NODE_KINDS_REVERSE = {
        pysvn.node_kind.none: "none",
        pysvn.node_kind.file: "file",
        pysvn.node_kind.dir:  "dir",
        pysvn.node_kind.unknown: "unknown"
    }

    def __init__(self):
        self.client = pysvn.Client()
        self.interface = "pysvn"
        self.vcs = rabbitvcs.vcs.VCS_SVN
        self.cache = rabbitvcs.vcs.status.StatusCache()

    def statuses(self, path, recurse=True, update=False, invalidate=False):
        """

        Look up the status for path.

        """
        
        if path in self.cache:
            if invalidate:
                del self.cache[path]
            else:
                return self.cache.find_path_statuses(path)

        on_error = rabbitvcs.vcs.status.Status.status_unknown(path)

        if not self.is_in_a_or_a_working_copy(path):
            return [on_error]

        depth = pysvn.depth.infinity if recurse else pysvn.depth.empty

        try:
            pysvn_statuses = self.client.status(path,
                                                depth=depth,
                                                update=update)
            if not len(pysvn_statuses):
                # This is NOT in the PySVN documentation, but sometimes it
                # returns an empty list if the file goes missing...
                return [on_error]
            else:
                statuslist = []
                for st in pysvn_statuses:
                    rabbitvcs_status = rabbitvcs.vcs.status.SVNStatus(st)
                    self.cache[st.path] = rabbitvcs_status
                    statuslist.append(rabbitvcs_status)
                    
                return statuslist
        except pysvn.ClientError, ex:
            # TODO: uncommenting these might not be a good idea
            #~ traceback.print_exc()
            log.debug("Exception occured in SVN.status() for %s" % path)
            log.exception(ex)
            return [on_error]

    def client_info(self, path):
        if islink(path):
            path = realpath(path)
        return self.client.info(path)

    def find_repository_path(self, path):
        path_to_check = path
        while path_to_check != "/" and path_to_check != "":
            if os.path.isdir(os.path.join(path_to_check, ".svn")):
                return path_to_check
            
            path_to_check = os.path.split(path_to_check)[0]
        
        return None

    def status(self, path, summarize=True, invalidate=False):
        if path in self.cache:
            if invalidate:
                del self.cache[path]
            else:
                st = self.cache[path]
                if summarize:
                    st.summary = st.single
                return st

        all_statuses = self.statuses(path, recurse=summarize)

        if summarize:
            path_status = None
            for st in all_statuses:
                if st.path == path:
                    path_status = st
                    break
            
            path_status.make_summary(all_statuses)
        else:
            path_status = all_statuses[0]

        return path_status

    def is_working_copy(self, path):
        try:
            # when a versioned directory is removed and replaced with a
            # non-versioned directory (one that doesn't have a working copy
            # administration area, or .svn directory) you can't do a status
            # call on that item itself (results in an exception).
            #
            # Note that this is not a conflict, it's more of a corruption.
            # And it's associated with the status "obstructed". The only
            # way to make sure that we're dealing with a working copy
            # is by verifying the SVN administration area exists.
            if (isdir(path) and
                    self.client_info(path) and
                    isdir(os.path.join(path, ".svn"))):
                return True
            return False
        except Exception, e:
            # FIXME: ClientError client in use on another thread
            #~ log.debug("EXCEPTION in is_working_copy(): %s" % str(e))
            return False

    def is_in_a_or_a_working_copy(self, path):
        if self.is_working_copy(path):
            return True

        return (self.find_repository_path(os.path.split(path)[0]) != "")

    def is_versioned(self, path):
        if self.is_working_copy(path):
            return True
        else:
            try:
                # info will return nothing for an unversioned file inside a working copy
                if (self.is_in_a_or_a_working_copy(path) and
                        self.client_info(path)):
                    return True
            except Exception, e:
                log.exception("is_versioned exception for %s" % path)
                
            return False

    def is_status(self, path, status_kind):
        try:
            status = self.status(path, summarize=False)
        except Exception, e:
            log.exception("is_status exception for %s" % path)
            return False

        # If looking for "NORMAL", then both statuses must be normal (or propstatus=none)
        # Otherwise, it is an either or situation
        if status_kind == pysvn.wc_status_kind.normal:
            return (status.data["text_status"] == status_kind
                and (status.data["prop_status"] == status_kind
                    or status.data["prop_status"] == pysvn.wc_status_kind.none))
        else:
            return (status.data["text_status"] == status_kind
                or status.data["prop_status"] == status_kind)

        return False

    def is_locked(self, path):
        is_locked = False
        try:
            is_locked = self.client.info2(path, recurse=False)[0][1].lock is not None
        except pysvn.ClientError, e:
            return False
            #log.exception("is_locked exception for %s" % path)

        return is_locked

    def get_items(self, paths, statuses=[]):
        """
        Retrieves a list of files that have one of a set of statuses

        @type   paths:      list
        @param  paths:      A list of paths or files.

        @type   statuses:   list
        @param  statuses:   A list of pysvn.wc_status_kind statuses.

        @rtype:             list
        @return:            A list of statuses

        """
        
        items = []

        for path in paths:

            sts = self.statuses(path, invalidate=True)
            for st in sts:
                if (not statuses) or (st.content in statuses or st.metadata in statuses):
                    items.append(st)

        return items

    def get_remote_updates(self, paths):
        if paths is None:
            return []
        
        items = []
        for path in paths:
            try:
                sts = self.statuses(path, update=True, invalidate=True)
            except Exception, e:
                log.exception(e)
                continue

            for st in sts:
                        
                if st.remote_content is None and st.remote_metadata is None:
                    continue

                if st.remote_content == "none" and st.remote_metadata == "none":
                    continue

                items.append(st)

        return items

    def get_repo_url(self, path):
        """
        Retrieve the repository URL for the given working copy path

        @type   path:   string
        @param  path:   A working copy path.

        @rtype:         string
        @return:        A repository URL.

        """

        # If the given path is a URL, the user is passing a repository url
        # In that case we already have the url
        if self.is_path_repository_url(path):
            return path

        # If the given path is not part of a working copy, keep trying the
        # parent path to see if it is part of a working copy
        path = self.get_versioned_path(os.path.abspath(path))
        if not path:
            return ""

        info = self.client_info(path)
        returner = ""
        try:
            returner = info["url"].encode('latin1')
        except Exception, e:
            log.exception(e)

        return returner

    def get_repo_root_url(self, path):
        """
        Retrieve the repository URL for the given working copy path
        FYI this method was not added until svn 1.6.x

        @type   path:   string
        @param  path:   A working copy path.

        @rtype:         string
        @return:        A repository URL.

        """

        info = self.client.info2(path, recurse=False)
        returner = ""
        try:
            returner = info[0][1]["repos_root_URL"]
        except Exception, e:
            log.exception(e)

        return returner

    def is_path_repository_url(self, path):
        for proto in ("http://", "https://", "svn://", "svn+ssh://", "file://"):
            if path.startswith(proto):
                return True

        return False

    def get_revision(self, path):
        """
        Retrieve the current revision number for a path

        @type   path:   string
        @param  path:   A working copy path.

        @rtype:         integer
        @return:        A repository revision.

        """

        info = self.client_info(path)

        returner = None
        try:
            returner = info["commit_revision"].number
        except KeyError, e:
            log.exception("KeyError exception in svn.py get_revision() for %s" % path)
        except AttributeError, e:
            log.exception("AttributeError exception in svn.py get_revision() for %s" % path)

        return returner

    def get_head(self, path):
        """
        Retrieve the HEAD revision for a repository.

        @type   path:   string
        @param  path:   A working copy path.

        @rtype:         integer
        @return:        A repository revision.

        """

        info = self.client_info(path)

        returner = None
        try:
            returner = info["revision"].number
        except KeyError, e:
            log.exception("KeyError exception in svn.py get_head() for %s" % path)
        except AttributeError, e:
            log.exception("AttributeError exception in svn.py get_head() for %s" % path)

        return returner

    def find_merge_candidate_revisions(self, from_url, to_path):
        """
        Return the list of revisions from from_url that are not already merged into to_path.

        @type  from_url:  string
        @param from_url:  A repository URL to merge from.

        @type  to_path:   string
        @param to_path:   A working copy path to merge into.

        @rtype:           list
        @return:          A list of revisions not already merged.
        """

        from_url_root = self.get_repo_root_url(from_url)
        from_branch = from_url.replace(from_url_root, '')
        from_branch = from_branch.rstrip('/')

        merge_info = self.propget(to_path, "svn:mergeinfo")

        merged_revisions = []
        for branch in merge_info.split('\n'):
            if not branch.startswith(from_branch + ':'):
                continue
            revisions = branch.split(':')[1]  # branch:rev,rev-rev,...
            for rev in revisions.split(','):
                if rev.find('-') != -1:
                    (rev_s, rev_e) = rev.split('-')
                    merged_revisions += range(int(rev_s), int(rev_e) + 1)
                else:
                    merged_revisions.append(int(rev))

        from_log = self.log(from_url, strict_node_history=False)
        from_revisions = [ int(l.revision.short()) for l in from_log ]

        to_log = self.log(to_path, strict_node_history=False)
        to_revisions = [ int(l.revision.short()) for l in to_log ]

        candidate_revisions = list(set(from_revisions) - set(to_revisions) - set(merged_revisions))

        return candidate_revisions


    #
    # properties
    #

    def get_versioned_path(self, path):
        """
        Generates a safe path to use with the prop* functions.
        If the given path is unversioned, go to the next path up.

        @type   path:   string
        @param  path:   A file or directory path.

        @rtype:         string
        @return:        A prop* function-safe path.

        """

        path_to_check = path
        path_to_use = None
        while path_to_check != "/" and path_to_check != "":
            if self.is_versioned(path_to_check):
                path_to_use = path_to_check
                return path_to_use

            path_to_check = os.path.split(path_to_check)[0]

        return path_to_use

    def propset(self, path, prop_name, prop_value, overwrite=False, recurse=False):
        """
        Adds an svn property to a path.  If the item is unversioned,
        add a recursive property to the parent path

        @type   path: string
        @param  path: A file or directory path.

        @type   prop_name: string
        @param  prop_name: An svn property name.

        @type   prop_value: string
        @param  prop_value: An svn property value/pattern.

        @type   recurse: boolean
        @param  recurse: If True, the property will be applied to all
                subdirectories as well.

        """

        path = self.get_versioned_path(path)
        if overwrite:
            props = prop_value
        else:
            props = self.propget(path, prop_name)
            props = "%s%s" % (props, prop_value)

        try:
            self.client.propset(
                prop_name,
                props,
                path,
                recurse=recurse
            )
            return True
        except pysvn.ClientError, e:
            log.exception("pysvn.ClientError exception in svn.py propset() for %s" % path)
        except TypeError, e:
            log.exception("TypeError exception in svn.py propset() %s" % path)

        return False

    def proplist(self, path, rev=None):
        """
        Retrieves a dictionary of properties for a path.

        @type   path:   string
        @param  path:   A file or directory path.

        @rtype:         dictionary
        @return:        A dictionary of properties.

        """
        if rev:
            returner = self.client.proplist(path, revision=rev)
        else:
            returner = self.client.proplist(path)
        if returner:
            returner = returner[0][1]
        else:
            returner = {}

        return returner

    def propget(self, path, prop_name, rev=None):
        """
        Retrieves a dictionary of the prop_value of the given
        path and prop_name

        @type   path:       string
        @param  path:       A file or directory path.

        @type   prop_name:  string or self.PROPERTIES
        @param  prop_name:  An svn property name.

        @rtype:             dictionary
        @return:            A dictionary where the key is the path, the value
                            is the prop_value.

        """

        path = self.get_versioned_path(path)
        try:
            if rev:
                returner = self.client.propget(
                    prop_name,
                    path,
                    recurse=True,
                    revision=rev
                )
            else:
                returner = self.client.propget(
                    prop_name,
                    path,
                    recurse=True
                )
        except pysvn.ClientError, e:
            log.exception("pysvn.ClientError exception in svn.py propget() for %s" % path)
            return ""

        try:
            returner = returner[path]
        except KeyError, e:
            returner = ""

        return returner

    def propdel(self, path, prop_name, recurse=True):
        """
        Removes a property from a given path

        @type   path: string
        @param  path: A file or directory path.

        @type   prop_name: string or self.PROPERTIES
        @param  prop_name: An svn property name.

        @type   recurse: boolean
        @param  recurse: If True, the property will be deleted from any
                subdirectories also having the property set.

        """

        path = self.get_versioned_path(path)

        returner = False
        try:
            self.client.propdel(
                prop_name,
                path,
                recurse=recurse
            )
            returner = True
        except pysvn.ClientError, e:
            log.exception("pysvn.ClientError exception in svn.py propdel() for %s" % path)
        except TypeError, e:
            log.exception("TypeError exception in svn.py propdel() %s" % path)

        return returner

    def propdetails(self, path):
        """
        Each property on a path may be modified in the WC, deleted or added.
        This method compares the properties on the local path to the base and
        identifies which.

        @param path: the path (file or dir) in the WC to check
        @type path: a path for something in a WC

        @return a dict of the form:
                {prop_name:
                    {"value": value (WC value, unless deleted then base value),
                     "status": status of property}
                }

        """
        local_props = self.proplist(path)
        base_props = self.proplist(path,
                                   rev=Revision("base").primitive())

        prop_details = {}

        local_propnames = set(local_props.keys())
        base_propnames = set(base_props.keys())

        for propname in (local_propnames | base_propnames):

            if propname in (local_propnames & base_propnames):
                # These are the property names that are common to the WC and
                # base. If their values have changed, list them as changed
                if local_props[propname] == base_props[propname]:
                    prop_details[propname] = {"status": rabbitvcs.vcs.status.status_normal,
                                              "value": local_props[propname]}

                else:
                    prop_details[propname] = {"status": rabbitvcs.vcs.status.status_modified,
                                              "value": local_props[propname]}

            elif propname in local_propnames:
                prop_details[propname] = {"status": rabbitvcs.vcs.status.status_added,
                                          "value": local_props[propname]}

            elif propname in base_propnames:
                prop_details[propname] = {"status": rabbitvcs.vcs.status.status_deleted,
                                          "value": base_props[propname]}

        return prop_details

    def revpropset(self, prop_name, prop_value, url, rev=None, force=False):
        """
        Adds an svn property to a path.  If the item is unversioned,
        add a recursive property to the parent path

        @type   url: string
        @param  url: A url to attach the prop to

        @type   prop_name: string
        @param  prop_name: An svn property name.

        @type   prop_value: string
        @param  prop_value: An svn property value/pattern.

        @type   rev: pysvn.Revision object
        @param  rev: The revision to attach the prop to

        @type   force: boolean
        @param  force: If True, the property will be forced

        """

        if rev is None:
            rev = self.revision("head")

        self.client.revpropset(prop_name, prop_value, url,
            revision=rev.primitive())

    def revproplist(self, url, rev=None):
        """
        Retrieves a dictionary of properties for a url.

        @type   url:   string
        @param  url:   A repository url

        @type   rev: pysvn.Revision object
        @param  rev: The revision to attach the prop to

        @rtype:         tuple(revision object, propsdict)
        @return:        A tuple with revision information and property dictionary

        """

        if rev is None:
            rev = self.revision("head")

        return self.client.revproplist(url, rev.primitive())[1]

    def revpropget(self, url, prop_name, rev=None):
        """
        Retrieves the revprop value for a specific url/propname/revision

        @type   url:       string
        @param  url:       A repository url

        @type   prop_name:  string or self.PROPERTIES
        @param  prop_name:  An svn property name.

        @type   rev: pysvn.Revision object
        @param  rev: The revision to attach the prop to

        @rtype:         tuple(revision object, propsdict)
        @return:        A tuple with revision information and property dictionary

        """

        if rev is None:
            rev = self.revision("head")

        return self.client.revpropget(
            prop_name,
            url,
            revision=rev.primitive()
        )

    def revpropdel(self, url, prop_name, rev=None, force=False):
        """
        Removes a property from a given path

        @type   url: string
        @param  url: A repository url

        @type   prop_name: string or self.PROPERTIES
        @param  prop_name: An svn property name.

        @type   rev: pysvn.Revision object
        @param  rev: The revision to attach the prop to

        @type   force: boolean
        @param  force: If True, the property deletion will be forced

        """

        if rev is None:
            rev = self.revision("head")

        return self.client.revpropdel(
            prop_name,
            url,
            revision=rev.primitive(),
            force=force
        )

    #
    # callbacks
    #

    def set_callback_cancel(self, func):
        self.client.callback_cancel = func

    def callback_cancel(self):
        if hasattr(self.client, "callback_cancel"):
            self.client.callback_cancel()

    def set_callback_notify(self, func):
        self.client.callback_notify = func

    def set_callback_get_log_message(self, func):
        self.client.callback_get_log_message = func

    def set_callback_get_login(self, func):
        self.client.callback_get_login = func

    def set_callback_ssl_server_trust_prompt(self, func):
        self.client.callback_ssl_server_trust_prompt = func

    def set_callback_ssl_client_cert_password_prompt(self, func):
        self.client.callback_ssl_client_cert_password_prompt = func

    def set_callback_ssl_client_cert_prompt(self, func):
        self.client.callback_ssl_client_cert_prompt = func

    #
    # revision
    #

    def revision(self, kind, date=None, number=None):
        """
        Create a revision object usable by pysvn

        @type   kind:   string
        @param  kind:   An svn.REVISIONS keyword.

        @type   date:   integer
        @param  date:   Used for kind=date, in the form of UNIX TIMESTAMP (secs).

        @type   number: integer
        @param  number: Used for kind=number, specifies the revision number.

        @rtype:         pysvn.Revision object
        @return:        A pysvn.Revision object.

        """

        # TODO: Don't use kwargs for date/number, just accept a "value" as a
        #       regular arg

        value = None
        if date:
            value = date
        elif number:
            value = number

        return Revision(kind, value)

    #
    # actions
    #

    def add(self, paths, recurse=True):
        """
        Add files or directories to the repository

        @type   paths: list
        @param  paths: A list of files/directories.

        @type   recurse: boolean
        @param  recurse: Recursively add a directory's children

        """

        return self.client.add(paths, recurse)

    def add_backwards(self, path):
        """
        This will add the given path to version control, and any parent
        directories that themselves require adding. It is essential that "path"
        contains a WC somewhere in its hierarchy.

        @param path: the path to add to version control
        @type path: string
        """
        head, tail = path,""
        tails = list()

        # We need to add backwards-recursively, since patch could create
        # files any level deep in the tree
        while not (self.is_working_copy(head) or self.is_versioned(head)):
            head, tail = os.path.split(head)
            tails.insert(0, tail)
            # If we get all the way to the FS root, something really dumb
            # has happened.
            assert head, "No longer in a working copy!"

        # Walk back up the tree...
        for tail in tails:
            head = os.path.join(head, tail)
            self.add(head, depth=pysvn.depth.empty)

    def copy(self, src, dest, revision=Revision("head")):
        """
        Copy files/directories from src to dest.  src or dest may both be either
        a local path or a repository URL.  revision is a pysvn.Revision object.

        @type   src: string
        @param  src: Source URL or path.

        @type   dest: string
        @param  dest: Destination URL or path.

        @type   revision: pysvn.Revision object
        @param  revision: A pysvn.Revision object.

        """
        src = rabbitvcs.util.helper.urlize(src)
        dest = rabbitvcs.util.helper.urlize(dest)
        return self.client.copy(src, dest, revision.primitive())

    def copy_all(self, sources, dest_url_or_path, copy_as_child=False,
            make_parents=False, ignore_externals=False):
        """
        Copy sources to the dest_url_or_path.

        @type   sources: list of tuples
        @param  sources: A list of tuples (url_or_path,revision)

        @type   dest_url_or_path: string
        @param  dest_url_or_path: Destination URL or path.

        @type   copy_as_child: boolean
        @param  copy_as_child: If there are multiple sources, copy as child
                    to dest_url_or_path (assumed to be a folder)

        @type   make_parents: boolean
        @param  make_parents: TBD

        @type   ignore_externals: boolean
        @param  ignore_externals: Omit externals

        """
        
        return self.client.copy2(sources, dest_url_or_path, copy_as_child,
            make_parents, None, ignore_externals)

    def checkout(self, url, path, recurse=True, revision=Revision("head"),
            ignore_externals=False):

        """
        Checkout a working copy from a vcs repository

        @type   url: string
        @param  url: A repository url.

        @type   path: string
        @param  path: A local destination for the working copy.

        @type   recurse: boolean
        @param  recurse: Whether or not to run a recursive checkout.

        @type   revision: pysvn.Revision
        @param  revision: Revision to checkout, defaults to HEAD.

        @type   ignore_externals: boolean
        @param  ignore_externals: Whether or not to ignore externals.

        """

        url = rabbitvcs.util.helper.urlize(url)

        return self.client.checkout(url, path, recurse=recurse,
            revision=revision.primitive(), ignore_externals=ignore_externals)

    def cleanup(self, path):
        """
        Clean up a working copy.

        @type   path: string
        @param  path: A local working copy path.

        """

        return self.client.cleanup(path)

    def revert(self, paths):
        """
        Revert files or directories so they are unversioned

        @type   paths: list
        @param  paths: A list of files/directories.

        """

        return self.client.revert(paths)

    def commit(self, paths, log_message="", recurse=False, keep_locks=False):
        """
        Commit a list of files to the repository.

        @type   paths: list
        @param  paths: A list of files/directories.

        @type   log_message: string
        @param  log_message: A commit log message.

        @type   recurse: boolean
        @param  recurse: Whether or not to recurse into sub-directories.

        @type   keep_locks: boolean
        @param  keep_locks: Whether or not to keep locks on commit.

        """
        
        kwargs = {"keep_locks": keep_locks}
        try:
            # Simply setting recurse=False will not stop child files from getting
            # committed.  The pysvn.depth kwarg must be set to empty.
            # Unfortunately, older pysvn/svn installations do not have the depth
            # enum so for those, recurse must be used.
            kwargs["depth"] = (recurse and pysvn.depth.infinity or pysvn.depth.empty)
        except AttributeError:
            kwargs["recurse"] = recurse
        
        retval = self.client.checkin(paths, log_message, **kwargs)
        dummy_commit_dict = {
            "revision": retval,
            "action": rabbitvcs.vcs.svn.commit_completed
            }
        self.client.callback_notify(dummy_commit_dict)

        return retval

    def log(self, url_or_path, revision_start=Revision("head"),
            revision_end=Revision("number", 0), limit=0,
            discover_changed_paths=True, strict_node_history=False):
        """
        Retrieve log items for a given path in the repository

        @type   url_or_path: string
        @param  url_or_path: Path for which to get log items for

        @type   revision_start: pysvn.Revision
        @param  revision_start: Most recent revision.  Defaults to HEAD

        @type   revision_end: pysvn.Revision
        @param  revision_end: Oldest revision.  Defaults to rev 0.

        @type   limit: int
        @param  limit: The maximum number of items to return.  Defaults to 0.

        """

        log = self.client.log(url_or_path, revision_start.primitive(),
            revision_end.primitive(), discover_changed_paths,
            strict_node_history, limit)

        returner = []
        for item in log:
            revision = Revision(pysvn.opt_revision_kind.number, item.revision.number)
            date = datetime.fromtimestamp(item.date)
            
            author = _("(no author)")
            if hasattr(item, "author"):
                author = item["author"]        

            message = ""
            if hasattr(item, "message"):
                message = item["message"]
        
            changed_paths = []
            for changed_path in item.changed_paths:
                copy_from_rev = ""
                if hasattr(changed_path.copyfrom_revision, "number"):
                    copy_from_rev = self.revision("number", changed_path.copyfrom_revision.number)
            
                copy_from_path = ""
                if hasattr(changed_path, "copy_from_path"):
                    copy_from_path = changed_path.copy_from_path
                    
                changed_paths.append(rabbitvcs.vcs.log.LogChangedPath(
                    changed_path.path,
                    changed_path.action,
                    copy_from_path,
                    copy_from_rev                    
                ))
        
            returner.append(rabbitvcs.vcs.log.Log(
                date,
                revision,
                author,
                message,
                changed_paths,
                None
            ))
        
        return returner

    def export(self, src_url_or_path, dest_path, revision=Revision("head"),
            recurse=True, ignore_externals=False, force=False, native_eol=None):

        """
        Export files from either a working copy or repository into a local
        path without versioning information.

        @type   src_url_or_path: string
        @param  src_url_or_path: A repository url.

        @type   dest_path: string
        @param  dest_path: A local destination for the working copy.

        @type   revision: pysvn.Revision
        @param  revision: The revision to retrieve from the repository.

        @type   ignore_externals: boolean
        @param  ignore_externals: Whether or not to ignore externals.

        @type   recurse: boolean
        @param  recurse: Whether or not to run a recursive checkout.

        """
        if os.path.exists(dest_path):
            if os.path.isdir(dest_path):
                shutil.rmtree(dest_path)
            else:
                os.remove(dest_path)



        return self.client.export(src_url_or_path, dest_path, force,
            revision.primitive(), native_eol, ignore_externals, recurse)

    def import_(self, path, url, log_message, ignore=False):

        """
        Import an unversioned file or directory structure into a repository.

        @type   path: string
        @param  path: An unversioned file or directory structure

        @type   url: string
        @param  url: A repository location to put the imported files

        @type   log_message: string
        @param  log_message: Log message to use for commit

        @type   ignore: boolean
        @param  ignore: Disregard svn:ignore props

        """

        url = rabbitvcs.util.helper.urlize(url)
        return self.client.import_(path, url, log_message, ignore)

    def lock(self, url_or_path, lock_comment, force=False):

        """
        Lock a url or path.

        @type   url_or_path: string
        @param  url_or_path: A url or path to lock

        @type   lock_comment: string
        @param  lock_comment: A log message to go along with the lock.

        @type   force: boolean
        @param  force: Steal the locks of others if they exist.

        """

        return self.client.lock(url_or_path, lock_comment, force)

    def relocate(self, from_url, to_url, path, recurse=True):

        """
        Relocate the working copy from from_url to to_url for path

        @type   from_url: string
        @param  from_url: A url to relocate from

        @type   to_url: string
        @param  to_url: A url to relocate to

        @type   path: string
        @param  path: The path of the local working copy

        """
        
        from_url = rabbitvcs.util.helper.urlize(from_url)
        to_url = rabbitvcs.util.helper.urlize(to_url)
        return self.client.relocate(from_url, to_url, path, recurse)

    def move(self, src_url_or_path, dest_url_or_path):

        """
        Schedule a file to be moved around the repository

        @type   src_url_or_path: string
        @param  src_url_or_path: A url/path to move from

        @type   dest_url_or_path: string
        @param  dest_url_or_path: A url/path to move to

        """

        if hasattr(self.client, "move2"):
            return self.client.move2([src_url_or_path], dest_url_or_path)
        else:
            return self.client.move(src_url_or_path, dest_url_or_path,
                force=True)

    def move_all(self, sources, dest_url_or_path, move_as_child=False,
            make_parents=False):
        """
        Move sources to the dest_url_or_path.

        @type   sources: list of tuples
        @param  sources: A list of tuples (url_or_path,revision)

        @type   dest_url_or_path: string
        @param  dest_url_or_path: Destination URL or path.

        @type   move_as_child: boolean
        @param  move_as_child: If there are multiple sources, move as child
                    to dest_url_or_path (assumed to be a folder)

        @type   make_parents: boolean
        @param  make_parents: TBD

        """

        return self.client.move2(sources, dest_url_or_path,
            move_as_child=move_as_child, make_parents=make_parents)

    def remove(self, url_or_path, force=False, keep_local=False):

        """
        Schedule a file to be removed from the repository

        @type   url_or_path: string
        @param  url_or_path: A url/path to remove

        @type   force: boolean
        @param  force: Force renaming, despite conflicts. Defaults to false.

        @type   keep_local: boolean
        @param  keep_local: Keep the local copy (don't just delete it)

        """

        return self.client.remove(url_or_path, force, keep_local)

    def revert(self, paths, recurse=False):
        """
        Revert files or directories from the repository

        @type   paths: list
        @param  paths: A list of files/directories.

        @type   recurse: boolean
        @param  recurse: Recursively add a directory's children

        """

        return self.client.revert(paths, recurse)

    def resolve(self, path, recurse=True):
        """
        Mark conflicted files as resolved

        @type   path: string
        @param  path: A local path to resolve

        @type   recurse: boolean
        @param  recurse: Recursively add a directory's children

        """

        return self.client.resolved(path, recurse)

    def switch(self, path, url, revision=Revision("head")):
        """
        Switch the working copy to another repository source.

        @type   path: string
        @param  path: A local path to a working copy

        @type   url: string
        @param  url: The repository location to switch to

        @type   revision: pysvn.Revision
        @param  revision: The revision of the repository to switch to (Def:HEAD)

        """

        url = rabbitvcs.util.helper.urlize(url)
        return self.client.switch(path, url, revision.primitive())

    def unlock(self, path, force=False):
        """
        Unlock locked files.

        @type   path: string
        @param  path: A local path to resolve

        @type   force: boolean
        @param  force: If locked by another user, unlock it anyway.

        """

        return self.client.unlock(path, force)

    def update(self, path, recurse=True, revision=Revision("head"),
            ignore_externals=False):
        """
        Update a working copy.

        @type   path: string
        @param  path: A local path to update

        @type   recurse: boolean
        @param  recurse: Update child folders recursively

        @type   revision: pysvn.Revision
        @param  revision: Revision to update to (Def: HEAD)

        @type   ignore_externals: boolean
        @param  ignore_externals: Ignore external items

        """

        return self.client.update(path, recurse, revision.primitive(),
            ignore_externals)

    def annotate(self, url_or_path, from_revision=Revision("number", 1),
            to_revision=Revision("head")):
        """
        Get the annotate results for the given file and revision range.

        @type   url_or_path: string
        @param  url_or_path: A url or local path

        @type   from_revision: pysvn.Revision
        @param  from_revision: Revision from (def: 1)

        @type   to_revision: pysvn.Revision
        @param  to_revision: Revision to (def: HEAD)

        """

        return self.client.annotate(url_or_path, from_revision.primitive(),
            to_revision.primitive())

    def merge_ranges(self, source, ranges_to_merge, peg_revision,
            target_wcpath, notice_ancestry=False, force=False, dry_run=False,
            record_only=False):
        """
        Merge a range of revisions.

        @type   source: string
        @param  source: A repository location

        @type   ranges_to_merge: list of tuples
        @param  ranges_to_merge: A list of revision ranges to merge

        @type   peg_revision: pysvn.Revision
        @param  peg_revision: Indicates which revision in sources is valid.

        @type   target_wcpath: string
        @param  target_wcpath: Target working copy path

        @type   notice_ancestry: boolean
        @param  notice_ancestry: unsure

        @type   force: boolean
        @param  force: unsure

        @type   dry_run: boolean
        @param  dry_run: Do a test/dry run or not

        @type   record_only: boolean
        @param  record_only: unsure

        TODO: Will firm up the parameter documentation later

        """
        return self.client.merge_peg2(source,
                                      ranges_to_merge,
                                      peg_revision.primitive(),
                                      target_wcpath,
                                      notice_ancestry=notice_ancestry,
                                      force=force,
                                      dry_run=dry_run,
                                      record_only=record_only)

    def has_merge2(self):
        """
        Tests whether the user has a later version of pysvn/svn installed
        with more merge features
        """
        return hasattr(self.client, "merge_peg2")

    def merge_trees(self, url_or_path1, revision1, url_or_path2, revision2,
            local_path, force=False, recurse=True, record_only=False):
        """
        Merge two trees into one.

        @type   url_or_path1: string
        @param  url_or_path1: From WC/URL location

        @type   revision1: pysvn.Revision
        @param  revision1: Indicates the revision of the URL/Path

        @type   url_or_path2: string
        @param  url_or_path2: To WC/URL location

        @type   revision2: pysvn.Revision
        @param  revision2: Indicates the revision of the URL/Path

        @type   local_path: string
        @param  local_path: Target working copy path

        @type   force: boolean
        @param  force: unsure

        @type   recurse: boolean
        @param  recurse: Merge children recursively

        @type   record_only: boolean
        @param  record_only: unsure

        TODO: Will firm up the parameter documentation later

        """

        url_or_path1 = rabbitvcs.util.helper.urlize(url_or_path1)
        url_or_path2 = rabbitvcs.util.helper.urlize(url_or_path2)
        return self.client.merge(url_or_path1, revision1.primitive(),
            url_or_path2, revision2.primitive(), local_path, force, recurse,
            record_only)

    def has_merge_reintegrate(self):
        """
        Tests whether the user has a later version of pysvn/svn installed
        with more merge features
        """
        return hasattr(self.client, "merge_reintegrate")

    def merge_reintegrate(self, url_or_path, revision, local_path, dry_run=False):
        """
        Merge a branch into trunk

        @type   url_or_path: string
        @param  url_or_path: From WC/URL location

        @type   revision: Revision()
        @param  revision: Indicates the revision of the URL/Path

        @type   local_path: string
        @param  local_path: Target working copy path

        @type   dry_run: boolean
        @param  dry_run: Do a test/dry run or not

        """

        return self.client.merge_reintegrate(url_or_path, revision.primitive(),
            local_path, dry_run)


    def diff(self, tmp_path, url_or_path, revision1, url_or_path2, revision2,
            recurse=True, ignore_ancestry=False, diff_deleted=True,
            ignore_content_type=False):
        """
        Returns the diff text between the base code and the working copy.

        @type   tmp_path: string
        @param  tmp_path: Temporal path to store the diff

        @type   url_or_path: string
        @param  url_or_path: From WC/URL location

        @type   revision1: pysvn.Revision
        @param  revision1: Indicates the revision of the URL/Path (def: pysvn.Revision( opt_revision_kind.base ))

        @type   url_or_path2: string
        @param  url_or_path2: From WC/URL location

        @type   revision2: pysvn.Revision
        @param  revision2: Indicates the revision of the URL/Path (def: pysvn.Revision( opt_revision_kind.working ))

        @type   recurse: boolean
        @param  recurse: Whether or not to recurse into sub-directories. (def: True)

        @type   ignore_ancestry: boolean
        @param  ignore_ancestry: Whether or not to recurse into sub-directories. (def: False)

        @type   diff_deleted: boolean
        @param  diff_deleted: Whether or not to recurse into sub-directories. (def: True)

        @type   ignore_content_type: boolean
        @param  ignore_content_type: Whether or not to recurse into sub-directories. (def: False)

        """

        return self.client.diff(tmp_path, url_or_path, revision1.primitive(),
            url_or_path2, revision2.primitive(), recurse, ignore_ancestry,
            diff_deleted, ignore_content_type)

    def diff_summarize(self, url_or_path1, revision1, url_or_path2, revision2,
            recurse=True, ignore_ancestry=False):
        """
        Returns a summary of changed items between two paths/revisions

        @type   url_or_path1: string
        @param  url_or_path1: First WC/URL location

        @type   revision1: pysvn.Revision
        @param  revision1: Indicates the revision of the URL/Path (def: pysvn.Revision( opt_revision_kind.base ))

        @type   url_or_path2: string
        @param  url_or_path2: Second WC/URL location

        @type   revision2: pysvn.Revision
        @param  revision2: Indicates the revision of the URL/Path (def: pysvn.Revision( opt_revision_kind.working ))

        @type   recurse: boolean
        @param  recurse: Whether or not to recurse into sub-directories. (def: True)

        @type   ignore_ancestry: boolean
        @param  ignore_ancestry: Whether or not to recurse into sub-directories. (def: False)

        @type   depth: pysvn.depth enum
        @param  depth: a replacement for recurse

        """

        return self.client.diff_summarize(url_or_path1, revision1.primitive(),
            url_or_path2, revision2.primitive(), recurse, ignore_ancestry)

    def list(self, url_or_path, revision=Revision("HEAD"), recurse=True):
        url_or_path = rabbitvcs.util.helper.urlize(url_or_path)
        return self.client.list(url_or_path, revision=revision.primitive(),
            recurse=recurse)

    def mkdir(self, url_or_path, log_message):
        """
        Make a new directory in the repository or working copy

        @type   url_or_path: string
        @param  url_or_path: Url in the repository or path in working copy

        @type   log_message: string
        @param  log_message: A log message to use in your commit

        """

        return self.client.mkdir(url_or_path, log_message)

    def apply_patch(self, patch_file, base_dir):
        """
        Applies a patch created for this WC.

        @type patch_file: string
        @param patch_file: the path to the patch file

        @type base_dir: string
        @param base_dir: the base directory from which to interpret the paths in
                         the patch file
        """

        any_failures = False

        for file, success, rej_file in rabbitvcs.util.helper.parse_patch_output(patch_file, base_dir):

            fullpath = os.path.join(base_dir, file)

            event_dict = dict()

            event_dict["path"] = file
            event_dict["mime_type"] = "" # meh

            if success:
                event_dict["action"] = _("Patched") # not in pysvn, but
                                                    # we have a fallback
            else:
                any_failures = True
                event_dict["action"] = _("Patch Failed") # better wording needed?

            # Creates its own notifications.
            self.add_backwards(fullpath)

            if rej_file:
                rej_info = {
                    "path" : rej_file,
                    "action" : _("Rejected Patch"),
                    "mime_type" : None
                            }

            if self.client.callback_notify:
                self.client.callback_notify(event_dict)
                if rej_file:
                    self.client.callback_notify(rej_info)



    def is_version_less_than(self, version):
        """
        @type   version: tuple
        @param  version: A version tuple to compare pysvn's version to
        """

        if version[0] > pysvn.version[0]:
            return True

        if ((version[0] == pysvn.version[0])
                and (version[1] > pysvn.version[1])):
            return True

        if ((version[0] == pysvn.version[0])
                and (version[1] == pysvn.version[1])
                and (version[2] > pysvn.version[2])):
            return True

        if ((version[0] == pysvn.version[0])
                and (version[1] == pysvn.version[1])
                and (version[2] == pysvn.version[2])
                and (version[3] > pysvn.version[3])):
            return True

        return False

    def is_version_greater_than(self, version):
        """
        @type   version: tuple
        @param  version: A version tuple to compare pysvn's version to
        """

        if version[0] < pysvn.version[0]:
            return True

        if ((version[0] == pysvn.version[0])
                and (version[1] < pysvn.version[1])):
            return True

        if ((version[0] == pysvn.version[0])
                and (version[1] == pysvn.version[1])
                and (version[2] < pysvn.version[2])):
            return True

        if ((version[0] == pysvn.version[0])
                and (version[1] == pysvn.version[1])
                and (version[2] == pysvn.version[2])
                and (version[3] < pysvn.version[3])):
            return True

        return False



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os.path

import pygtk
import gobject
import gtk
import os
import commands

from rabbitvcs.ui import InterfaceNonView
from rabbitvcs.ui.action import SVNAction
import rabbitvcs.vcs
from rabbitvcs.util.log import Log

log = Log("rabbitvcs.ui.applypatch")

from rabbitvcs import gettext
_ = gettext.gettext

class ApplyPatch(InterfaceNonView):
    """
    This class provides a handler to the apply patch functionality.
    
    """

    def __init__(self, paths):
        InterfaceNonView.__init__(self)
        self.paths = paths
        self.vcs = rabbitvcs.vcs.VCS()
        self.common = rabbitvcs.util.helper.get_common_directory(paths)

    def choose_patch_path(self):
        path = None
        
        dialog = gtk.FileChooserDialog(
            _("Apply Patch"),
            None,
            gtk.FILE_CHOOSER_ACTION_OPEN,(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                          gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            path = dialog.get_filename()
        
        dialog.destroy()
        
        return path
    
    def choose_patch_dir(self):
        if len(self.paths) == 1 and os.path.isdir(self.paths[0]):
            return self.paths[0]
        
        dir = None
        
        dialog = gtk.FileChooserDialog(
                    _("Apply Patch To Directory..."),
                    None,
                    gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                    (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                     gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        
        response = dialog.run()
        
        if response == gtk.RESPONSE_OK:
            dir = dialog.get_filename()
            
        dialog.destroy()
        
        return dir

class SVNApplyPatch(ApplyPatch):
    def __init__(self, paths):
        ApplyPatch.__init__(self, paths)
        
        self.svn = self.vcs.svn()

    def start(self):
    
        path = self.choose_patch_path()
        # If empty path, means we've cancelled
        if not path:
            return
        
        base_dir = self.choose_patch_dir()
        if not base_dir:
            return        
        
        ticks = 2
        self.action = rabbitvcs.ui.action.SVNAction(
            self.svn,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        self.action.set_pbar_ticks(ticks)
        self.action.append(self.action.set_header, _("Apply Patch"))
        self.action.append(self.action.set_status, _("Applying Patch File..."))
        self.action.append(self.svn.apply_patch, path, base_dir)
        self.action.append(self.action.set_status, _("Patch File Applied"))
        self.action.append(self.action.finish)
        self.action.start()

class GitApplyPatch(ApplyPatch):
    def __init__(self, paths):
        ApplyPatch.__init__(self, paths)
        
        self.git = self.vcs.git(paths[0])

    def start(self):
    
        path = self.choose_patch_path()
        # If empty path, means we've cancelled
        if not path:
            return
        
        base_dir = self.choose_patch_dir()
        if not base_dir:
            return        
        
        ticks = 2
        self.action = rabbitvcs.ui.action.GitAction(
            self.git,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        self.action.set_pbar_ticks(ticks)
        self.action.append(self.action.set_header, _("Apply Patch"))
        self.action.append(self.action.set_status, _("Applying Patch File..."))
        self.action.append(self.git.apply_patch, path, base_dir)
        self.action.append(self.action.set_status, _("Patch File Applied"))
        self.action.append(self.action.finish)
        self.action.start()

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNApplyPatch,
    rabbitvcs.vcs.VCS_GIT: GitApplyPatch
}

def applypatch_factory(paths):
    guess = rabbitvcs.vcs.guess(paths[0])
    return classes_map[guess["vcs"]](paths)

if __name__ == "__main__":
    from rabbitvcs.ui import main
    (options, paths) = main(usage="Usage: rabbitvcs applypatch [path1] [path2] ...")

    window = applypatch_factory(paths)
    window.register_gtk_quit()
    window.start()
    gtk.main()
    



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os.path
import urllib

import pygtk
import gobject
import gtk

from rabbitvcs.ui import InterfaceView
from rabbitvcs.ui.checkout import Checkout
import rabbitvcs.ui.widget
import rabbitvcs.ui.dialog
import rabbitvcs.ui.action
import rabbitvcs.util.helper
import rabbitvcs.vcs

from rabbitvcs import gettext
_ = gettext.gettext

class GitClone(Checkout):
    def __init__(self, path=None, url=None):
        Checkout.__init__(self, path, url)
        
        self.git = self.vcs.git()
        
        self.get_widget("Checkout").set_title(_("Clone"))
        self.get_widget("repo_chooser").hide()

        self.check_form()

    def on_ok_clicked(self, widget):
        url = self.repositories.get_active_text().strip()
        path = self._get_path().strip()
        
        if not url or not path:
            rabbitvcs.ui.dialog.MessageBox(_("The repository URL and destination path are both required fields."))
            return
    
        self.hide()
        self.action = rabbitvcs.ui.action.GitAction(
            self.git,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        self.action.append(self.action.set_header, _("Clone"))
        self.action.append(self.action.set_status, _("Running Clone Command..."))
        self.action.append(rabbitvcs.util.helper.save_repository_path, url)
        self.action.append(
            self.git.clone,
            url,
            path
        )
        self.action.append(self.action.set_status, _("Completed Clone"))
        self.action.append(self.action.finish)
        self.action.start()

    def on_repositories_changed(self, widget, data=None):
        url = self.repositories.get_active_text()
        tmp = url.replace("//", "/").split("/")[1:]
        append = ""
        prev = ""
        while len(tmp):
            prev = append
            append = tmp.pop()
                
            if append in ("http:", "https:", "file:", "git:"):
                append = ""
                break

            if append.endswith(".git"):
                append = append[:-4]
                break

        self.get_widget("destination").set_text(
            os.path.join(self.destination, append)
        )
        
        self.check_form()

    def check_form(self):
        self.complete = True
        if self.repositories.get_active_text() == "":
            self.complete = False
        if self.get_widget("destination").get_text() == "":
            self.complete = False
        
        self.get_widget("ok").set_sensitive(self.complete)

classes_map = {
    rabbitvcs.vcs.VCS_GIT: GitClone
}

def clone_factory(classes_map, vcs, path=None, url=None):
    return classes_map[vcs](path, url)

if __name__ == "__main__":
    from rabbitvcs.ui import main, VCS_OPT
    (options, args) = main(
        [VCS_OPT],
        usage="Usage: rabbitvcs clone --vcs=git [url] [path]"
    )
    
    # Default to using git
    vcs = rabbitvcs.vcs.VCS_GIT
    if options.vcs:
        vcs = options.vcs
    
    # If two arguments are passed:
    #   The first argument is expected to be a url
    #   The second argument is expected to be a path
    # If one argument is passed:
    #   If the argument exists, it is a path
    #   Otherwise, it is a url
    path = url = None
    if len(args) == 2:
        path = args[0]
        url = args[1]
    elif len(args) == 1:
        if os.path.exists(args[0]):
            path = args[0]
        else:
            url = args[0]

    window = clone_factory(classes_map, vcs, path=path, url=url)
    window.register_gtk_quit()
    gtk.main()



#
# test/stage.py
#

import os
from shutil import rmtree
from sys import argv
from optparse import OptionParser

from gittyup.client import GittyupClient
from util import touch

parser = OptionParser()
parser.add_option("-c", "--cleanup", action="store_true", default=False)
(options, args) = parser.parse_args(argv)

DIR = "branch"

if options.cleanup:
    rmtree(DIR, ignore_errors=True)

    print "branch.py clean"
else:
    if os.path.isdir(DIR):
        raise SystemExit("This test script has already been run.  Please call this script with --cleanup to start again")

    os.mkdir(DIR)
    g = GittyupClient()
    g.initialize_repository(DIR)
    
    touch(DIR + "/test1.txt")
    touch(DIR + "/test2.txt")
    
    g.stage([DIR+"/test1.txt", DIR+"/test2.txt"])
    g.commit("This is a commit")

    # Create a new branch, don't track it
    g.branch("branch1")
    assert ("branch1" in g.branch_list())

    # Make sure we are still tracking master
    assert (g.is_tracking("refs/heads/master"))

    # Track branch1
    g.track("refs/heads/branch1")
    assert (g.is_tracking("refs/heads/branch1"))
    
    # Rename branch1 to branch1b
    g.branch_rename("branch1", "branch1b")
    assert ("branch1b" in g.branch_list())

    # Make sure we are now tracking branch1b
    assert (g.is_tracking("refs/heads/branch1b"))
    
    # Delete branch1b
    g.branch_delete("branch1b")
    assert ("branch1b" not in g.branch_list())

    # Make sure we are now tracking master
    assert (g.is_tracking("refs/heads/master"))

    print "branch.py pass"



import locale
import os

from rabbitvcs.util.log import Log
import rabbitvcs.util.settings
import rabbitvcs.util.helper

log = Log("rabbitvcs.util.locale")

def initialize_locale():
    try:
        settings = rabbitvcs.util.settings.SettingsManager()

        sane_default = locale.getdefaultlocale(['LANG', 'LANGUAGE'])

        # Just try to set the default locale for the user
        locale.setlocale(locale.LC_ALL, sane_default)

        # Now, if the user has set a default, try to apply that
        user_default = settings.get("general", "language")
        if user_default:
            locale.setlocale(locale.LC_ALL, (user_default, sane_default[1]))
    except locale.Error:
        # If the user's environment does not specify an encoding, Python will
        # pick a default which might not be available. It seems to pick
        # ISO8859-1 (latin1), but UTF8 is a better idea on GNU/Linux.
        log.warning("Could not set default locale (LANG: %s)" % os.environ.get("LANG"))
        
        (loc, enc) = sane_default
        
        # We should only try this if we have a region to set as well.
        if loc and enc != "UTF8":
            try:
                locale.setlocale(locale.LC_ALL, (loc, "UTF8"))
                log.warning("Manually set encoding to UTF-8")
            except locale.Error:
                # Nope, no UTF8 either.
                log.warning("Could not set user's locale to UTF-8")



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

try:
    from gi.repository import GObject as gobject
except ImportError:
    import gobject

import os
if "NAUTILUS_PYTHON_REQUIRE_GTK3" in os.environ and os.environ["NAUTILUS_PYTHON_REQUIRE_GTK3"]:
    from gi.repository import Gtk as gtk
    GTK_FILL = gtk.AttachOptions.FILL
    GTK_EXPAND = gtk.AttachOptions.EXPAND
    GTK3 = True
else:
    import gtk
    GTK_FILL = gtk.FILL
    GTK_EXPAND = gtk.EXPAND
    GTK3 = False

# GI Pango is broken in some versions of pygobject 2.28.x, if that is the case
# set HAS_PANGO to false and do without ellipsizing for now
HAS_PANGO = True
try:
    from gi.repository import Pango as pango
    PANGO_ELLIPSIZE_MIDDLE = pango.EllipsizeMode.MIDDLE
    PANGO_ELLIPSIZE_END = pango.EllipsizeMode.END
except ImportError:
    import pango
    PANGO_ELLIPSIZE_MIDDLE = pango.ELLIPSIZE_MIDDLE
    PANGO_ELLIPSIZE_END = pango.ELLIPSIZE_END
except AttributeError:
    HAS_PANGO = False
    pass


import os.path

HAS_GTKSPELL = False
if not GTK3:
    try:
        import gtkspell
        HAS_GTKSPELL = True
    except ImportError:
        pass    

HAS_GTKSOURCEVIEW = False
if not GTK3:
    try:
        import gtksourceview
        HAS_GTKSOURCEVIEW = True
    except ImportError:
        pass

import rabbitvcs.util.helper

from rabbitvcs import gettext
_ = gettext.gettext

from rabbitvcs.util.log import Log
log = Log("rabbitvcs.ui.widget")

from rabbitvcs.ui import STATUS_EMBLEMS

TOGGLE_BUTTON = 'TOGGLE_BUTTON'
TYPE_PATH = 'TYPE_PATH'
TYPE_STATUS = 'TYPE_STATUS'
TYPE_ELLIPSIZED = 'TYPE_ELLIPSIZED'
TYPE_GRAPH = 'TYPE_GRAPH'
TYPE_MARKUP = 'TYPE_MARKUP'
TYPE_HIDDEN = 'TYPE_HIDDEN'

ELLIPSIZE_COLUMN_CHARS = 20

PATH_ENTRY = 'PATH_ENTRY'
SEPARATOR = u'\u2015' * 10

from pprint import pformat
def filter_router(model, iter, column, filters):
    """
    Route filter requests for a table's columns.  This function is called for
    each cell of the table that gets displayed.
    
    @type   model: gtk.TreeModelFilter or gtk.TreeModelSort
    @param  model: The TreeModelFilter or gtk.TreeModelSort instance for our
                   table
    
    @type   iter: gtk.TreeIter
    @param  iter: The TreeIter instance for the table row being filtered
    
    @type   column: int
    @param  column: The column index of the current item being filtered
    
    @type   filters: list
    @param  filters: A list of dicts used to define how a column should be
        filtered
        
        Note for filters:  Each dict consists of a callback function and user
        data like so:
        
            {
                "callback": self.file_filter,
                "user_data": {
                    "column": 0, //tells the callback what column to filter
                    "base_dir": "/home/workingcopy"
                }
            }
    
    @return    The filtered output defined for the given column
    
    """
    real_model = model.get_model()
    
    real_iter = model.convert_iter_to_child_iter(iter)
    
    row = real_model[real_model.get_path(real_iter)]

    if not filters:
        return row[column]

    for filter in filters:
        filter_column = filter["user_data"]["column"]
        if column == filter_column:
            return filter["callback"](row, column, filter["user_data"])

    return row[column]

def path_filter(row, column, user_data=None):
    """
    A common filter function that is used in many tables.  Changes the displayed
    path to a path relative to the given base_dir (current working directory)
    
    @type   row: gtk.TreeModelRow
    @param  row: The row that is being filtered
    
    @type   column: int
    @param  column: The column that is being filtered
    
    @type   user_data: dict
    @param  user_data: A dictionary of user_data useful to this function
    
    @rtype  str
    @return A relative path
    
    """
    base_dir = user_data["base_dir"]

    if row[column]:
        relpath = rabbitvcs.util.helper.get_relative_path(base_dir, row[column])
        if relpath == "":
            relpath = os.path.basename(row[column])
        return relpath
    else:
        return row[column] 

def long_text_filter(row, column, user_data=None):
    """
    Uses the format_long_text helper function to trim and prettify some text.
    """
    text = row[column]
    
    cols = user_data["cols"]
    
    
    if text:
        text = rabbitvcs.util.helper.format_long_text(text, cols)
        
    return text

def git_revision_filter(row, column, user_data=None):
    """
    Only show the first six characters of a git revision hash
    """
    
    text = row[column]
    if text:
        if text.startswith("<b>"):
            text = text[3:10]
        else:
            text = text[0:7]

    return text

def translate_filter(row, column, user_data=None):
    """
    Translates text as needed.
    """
    text = row[column]
    if text: return _(text)
    
def compare_items(model, iter1, iter2, user_data=None):

    if not user_data:
        # No column data given => Give up
        return 0

    colnum, coltype = user_data
    
    real_model = model.get_model()
    real_iter1 = model.convert_iter_to_child_iter(iter1)
    real_iter2 = model.convert_iter_to_child_iter(iter2)
    
    value1 = real_model.get_value(real_iter1, colnum)
    value2 = real_model.get_value(real_iter2, colnum)
    
    if value1 == value2:
        return 0
    elif value1 < value2:
        return -1
    else:
        return 1
    
class TableBase:
    def __init__(self, treeview, coltypes, colnames, values=[], filters=None,
                 filter_types=None, callbacks={}, flags={}):
        """
        @type   treeview: gtk.Treeview
        @param  treeview: The treeview widget to use
        
        @type   coltypes: list
        @param  coltypes: Contains the "type" of each column (i.e. str or int)
        
        @type   colnames: list
        @param  colnames: Contains the name string for each column
        
        @type   values: list
        @param  values: Contains the data to be inserted into the table
        
        @type   filters: list
        @param  filters: A list of dicts used to define how a column should be
            filtered
            
            Note for filters:  Each dict consists of a callback function and user
            data like so:
            
                {
                    "callback": self.file_filter,
                    "user_data": {
                        "column": 0, //tells the callback what column to filter
                        "base_dir": "/home/workingcopy"
                    }
                }
        
        @type   filter_types: list
        @param  filter_types: Contains the filtered "type" of each column.
        
        @type   callbacks: dict
        @param  callbacks: A dict of callbacks to be used.  Some are for signal
            handling while others are useful for other things.
            
        @type   flags: dict
        @param  flags: A dict of flags
        
        FLAGS:
            @type   sortable: boolean
            @param  sortable: whether the columns can be sorted
            
            @type   sort_on: int
            @param  sort_on: the column number to initially sort by
            
            @type   editable: list
            @param  editable: A list of which columns are editable
        
        """
        
        from renderers.graphcell import CellRendererGraph

        if not flags.has_key("sortable"):
            flags["sortable"] = False
        if not flags.has_key("sort_on"):
            flags["sort_on"] = -1
        if not flags.has_key("editable"):
            flags["editable"] = ()
    
        self.treeview = treeview
        self.selected_rows = []

        i = 0       
        for name in colnames:
            if coltypes[i] == gobject.TYPE_BOOLEAN:
                cell = gtk.CellRendererToggle()
                cell.set_property('activatable', True)
                cell.set_property('xalign', 0)
                cell.connect("toggled", self.toggled_cb, i)
                
                colname = ""
                if name != TOGGLE_BUTTON:
                    colname = name
                
                col = gtk.TreeViewColumn(colname, cell)
                col.set_attributes(cell, active=i)
            elif coltypes[i] == TYPE_PATH:
                # The type should be str but we have to use TYPE_PATH to
                # distinguish from a regular str column
                coltypes[i] = str
                
                # First we create the column, then we create a CellRenderer 
                # instance for the path icon and a CellRenderer instance for
                # the path.  Each is packed into the treeview column
                col = gtk.TreeViewColumn(name)

                cellpb = gtk.CellRendererPixbuf()
                cellpb.set_property('xalign', 0)
                cellpb.set_property('yalign', 0)
                col.pack_start(cellpb, False)
                data = None
                if callbacks.has_key("file-column-callback"):
                    data = {
                        "callback": callbacks["file-column-callback"],
                        "column": i
                    }
                else:
                    data = {
                        "callback": rabbitvcs.util.helper.get_node_kind,
                        "column": i
                    }
                col.set_cell_data_func(cellpb, self.file_pixbuf, data)
                
                cell = gtk.CellRendererText()
                cell.set_property('xalign', 0)
                cell.set_property('yalign', 0)
                col.pack_start(cell, False)
                col.set_attributes(cell, text=i)
            elif coltypes[i] == TYPE_STATUS:
                # Same as for TYPE_PATH
                coltypes[i] = str                
                col = gtk.TreeViewColumn(name)
                
                cellpb = gtk.CellRendererPixbuf()
                cellpb.set_property('xalign', 0)
                cellpb.set_property('yalign', 0)
                
                col.pack_start(cellpb, False)
                
                data = None
                
                col.set_cell_data_func(cellpb, self.status_pixbuf, i)
                
                cell = gtk.CellRendererText()
                cell.set_property('xalign', 0)
                cell.set_property('yalign', 0)
                col.pack_start(cell, False)
                col.set_attributes(cell, text=i)
            elif coltypes[i] == TYPE_HIDDEN:
                coltypes[i] = str
                col = gtk.TreeViewColumn(name, cell)
                col.set_visible(False)
            elif coltypes[i] == TYPE_ELLIPSIZED:
                coltypes[i] = str
                cell = gtk.CellRendererText()
                cell.set_property('yalign', 0)
                cell.set_property('xalign', 0)
                if HAS_PANGO:
                    cell.set_property('ellipsize', PANGO_ELLIPSIZE_END)
                    cell.set_property('width-chars', ELLIPSIZE_COLUMN_CHARS)
                col = gtk.TreeViewColumn(name, cell)
                col.set_attributes(cell, text=i)
            elif coltypes[i] == TYPE_GRAPH:
                coltypes[i] = gobject.TYPE_PYOBJECT
                cell = CellRendererGraph()
                col = gtk.TreeViewColumn(name, cell)
                col.add_attribute(cell, "graph", i)
            else:
                cell = gtk.CellRendererText()
                cell.set_property('yalign', 0)
                cell.set_property('xalign', 0)
                if i in flags["editable"]:
                    cell.set_property('editable', True)
                    cell.connect("edited", self.__cell_edited, i)
                
                if coltypes[i] == TYPE_MARKUP:
                    col = gtk.TreeViewColumn(name, cell, markup=i)
                else:
                    col = gtk.TreeViewColumn(name, cell, text=i)

                coltypes[i] = gobject.TYPE_STRING

            if flags["sortable"]:
                col.set_sort_column_id(i)

            self.treeview.append_column(col)
            i += 1

        self.data = self.get_store(coltypes)

        self.sorted = None

        # self.sorted == sorted view of data
        # self.filter == filtered data (abs paths -> rel paths)
        # self.data == actual data

        # The filter is there to change the way data is displayed. The data
        # should always be accessed via self.data, NOT self.filter.
        self.filter = self.data.filter_new()
        types = (filter_types and filter_types or coltypes)
        self.filter.set_modify_func(
                        types,
                        filter_router,
                        filters)


		# This runs through the columns, and sets the "compare_items" comparator
        # as needed. Note that the user data tells which column to sort on.
        if flags["sortable"]:
            self.sorted = gtk.TreeModelSort(self.filter)
            
            self.sorted.set_default_sort_func(compare_items, None)
            
            for idx in range(0, i):
                self.sorted.set_sort_func(idx,
                                          compare_items,
                                          (idx, coltypes[idx]))
               
            self.sorted.set_sort_column_id(flags["sort_on"], gtk.SORT_ASCENDING)
            
            self.treeview.set_model(self.sorted)
            
        elif filters:
            self.treeview.set_model(self.filter)
        else:
            self.treeview.set_model(self.data)

        
        if len(values) > 0:
            self.populate(values)
    
        self.set_resizable()

        # Set up some callbacks for all tables to deal with row clicking and
        # selctions
        self.treeview.connect("cursor-changed", self.__cursor_changed_event)
        self.treeview.connect("row-activated", self.__row_activated_event)
        self.treeview.connect("button-press-event", self.__button_press_event)
        self.treeview.connect("button-release-event", self.__button_release_event)
        self.treeview.connect("key-press-event", self.__key_press_event)
        self.treeview.connect("select-cursor-row", self.__row_selected)
        self.callbacks = callbacks
        if self.callbacks:
            self.allow_multiple()

    def _realpath(self, visible_path):
        """
        Converts a path (ie. row number) that we get from what the user selects
        into a path for the underlying data structure. If the data is not
        sorted, this is trivial; if it is sorted, the sorting model can figure
        it out for us.
        """
        if self.sorted:
            path = self.sorted.convert_path_to_child_path(visible_path)
        else:
            path = self.filter.convert_path_to_child_path(visible_path)
        return path

    def toggled_cb(self, cell, path, column):
        model = self.data
        realpath = self._realpath(path)
        model[realpath][column] = not model[realpath][column]
        if "row-toggled" in self.callbacks:
            self.callbacks["row-toggled"](model[realpath], column)


    def append(self, row):
        self.data.append(row)

    def remove(self, index):
        model = self.data
        del model[index]

    def remove_multiple(self, rows):
        i = 0
        for row in rows:
            rm_index = row
            if i > 0:
                rm_index -= 1
            
            self.remove(rm_index)
            i += 1     

    def get_items(self):
        return self.data

    def clear(self):
        self.data.clear()
        self.reset_selection()
        
    def get_row(self, index):
        model = self.data
        return model[index]
    
    def set_row(self, index, row):
        model = self.data
        model[index] = row
    
    def set_row_item(self, row, column, val):
        model = self.data
        model[row][column] = val
    
    def allow_multiple(self):
        self.treeview.get_selection().set_mode(gtk.SELECTION_MULTIPLE)

    def get_activated_rows(self, column=None):
        returner = []
        for row in self.data:
            if row[0]:
                item = row
                if column is not None:
                    item = row[column]
                
                returner.append(item)
                
        return returner
    
    def scroll_to_bottom(self):
        bottom = len(self.get_items()) - 1
        self.treeview.scroll_to_cell(bottom)

    def set_resizable(self, resizable=True):
        for col in self.treeview.get_columns():
            col.set_resizable(resizable)

    def get_column(self, column):
        return self.treeview.get_column(column)

    def set_column_width(self, column, width=None):
        col = self.treeview.get_column(column)
        if width is not None:
            col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
            col.set_fixed_width(width)
        else:
            col.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)

    def update_selection(self):
        selection = self.treeview.get_selection()
        (liststore, indexes) = selection.get_selected_rows()

        self.reset_selection()
        
        for tup in indexes:
            self.selected_rows.append(self._realpath(tup)[0])

    def reset_selection(self):
        self.selected_rows = []

    def get_selected_row_items(self, col):
        items = []
        for row in self.selected_rows:
            items.append(self.data[row][col])
        
        return items

    def get_selected_rows(self):
        return self.selected_rows
        
    def generate_string_from_data(self):
        lines = []
        for row in self.data:
            line = []
            for cell in row:
                line.append(unicode(cell))
            lines.append("\t".join(line))
        
        return "\n".join(lines)
        
    def unselect_all(self):
        self.treeview.get_selection().unselect_all()

    def focus(self, row, column):
        treecol = self.treeview.get_column(column)
        self.treeview.set_cursor((row,), treecol)
        self.treeview.grab_focus()

    def __button_press_event(self, treeview, data):
        info = treeview.get_path_at_pos(int(data.x), int(data.y))
        selection = treeview.get_selection()
        
        # If info is none, that means the user is clicking the empty space
        # In that case, unselect everything and update the selected_rows
        if info is None:
            selection.unselect_all()
            self.update_selection()
            return
            
        # this allows us to retain multiple selections with a right-click
        if data.button == 3:
            (liststore, indexes) = selection.get_selected_rows()
            
            # If the mouse click is one of the currently selected rows
            # keep the selection, otherwise, use the new selection
            for index in indexes:
                if index[0] == info[0][0]:
                    return True
            return False

    def __row_activated_event(self, treeview, data, col):
        treeview.grab_focus()
        self.update_selection()
        if "row-activated" in self.callbacks:
            self.callbacks["row-activated"](treeview, data, col)
    
    def __row_selected(self, treeview, started_editing):
        self.update_selection()
        if "row-selected" in self.callbacks:
            self.callbacks["row-selected"](treeview, started_editing)
        
    def __key_press_event(self, treeview, data):
        self.update_selection()
        if "key-event" in self.callbacks:
            self.callbacks["key-event"](treeview, data)

    def __cursor_changed_event(self, treeview):
        self.update_selection()
        if "cursor-changed" in self.callbacks:
            self.callbacks["cursor-changed"](treeview)
        if "mouse-event" in self.callbacks:
            self.callbacks["mouse-event"](treeview)

    def __button_release_event(self, treeview, data):
        self.update_selection()
        if "mouse-event" in self.callbacks:
            self.callbacks["mouse-event"](treeview, data)

    def __cell_edited(self, cell, row, data, column):
        self.update_selection()
        if "cell-edited" in self.callbacks:
            self.callbacks["cell-edited"](cell, row, data, column)

#    def __column_header_clicked(self, column, column_idx):
#        self.data.set_sort_column_id(column_idx, )

    def status_pixbuf(self, column, cell, model, iter, colnum):
        
        path = model.get_path(iter)
    
        real_path = model.convert_path_to_child_path(path)
        
        status = self.data[real_path][colnum]
        
        if status not in STATUS_EMBLEMS.keys():
            status = "error"
             
        icon = "emblem-" + STATUS_EMBLEMS[status]
        
        cell.set_property("icon_name", icon)
        

    def file_pixbuf(self, column, cell, model, iter, data=None):
        stock_id = None
        
        path = model.get_path(iter)
    
        real_path = model.convert_path_to_child_path(path)
            
        if data:
            real_item = self.data[real_path][data["column"]]
            kind = data["callback"](real_item)
            stock_id = gtk.STOCK_FILE
            if kind == "dir":
                stock_id = gtk.STOCK_DIRECTORY

        if stock_id is not None:
            cell.set_property("stock_id", stock_id)

class Table(TableBase):
    """
    Generate a flat tree view.
        
    See the TableBase documentation for parameter information

    """
    
    def __init__(self, treeview, coltypes, colnames, values=[], filters=None, 
            filter_types=None, callbacks={}, flags={}):
        TableBase.__init__(self, treeview, coltypes, colnames, values, filters, 
            filter_types, callbacks, flags)
    
    def get_store(self, coltypes):
        return gtk.ListStore(*coltypes)

    def populate(self, values):
        for row in values:
            self.data.append(row)

class Tree(TableBase):
    """
    Generate a hierarchal tree view.  The structure of "values" should be like:

        values = [
            (["A"], [
                (["C"], None)
            ]),
            (["B"], [
                (["D"], [
                    (["E"], None)
                ])
            ])
        ]
        
        Note that with multiple columns, you add to the list in the first element
        of each tuple.  (i.e. ["A"] becomes ["A", "Z", ... ]
        
    See the TableBase documentation for parameter information

    """
    def __init__(self, treeview, coltypes, colnames, values=[], filters=None, 
            filter_types=None, callbacks={}, flags={}):
        TableBase.__init__(self, treeview, coltypes, colnames, values, filters, 
            filter_types, callbacks, flags={})
    
    def get_store(self, coltypes):
        return gtk.TreeStore(*coltypes)

    def populate(self, values, parent=None):
        for node in values:
            root = node[0]
            new_root = self.data.append(parent, root)
            if len(node) > 1 and node[1] is not None:
                self.populate(node[1], new_root)
        
class ComboBox:
    def __init__(self, cb, items=None):
    
        self.cb = cb
    
        self.model = gtk.ListStore(str)
        if items is not None:
            for i in items:
                self.append(i)

        self.cb.set_model(self.model)

        if type(self.cb) == gtk.ComboBoxEntry:
            self.cb.set_text_column(0)
        elif type(self.cb) == gtk.ComboBox:
            cell = gtk.CellRendererText()
            self.cb.pack_start(cell, True)
            self.cb.add_attribute(cell, 'text', 0)

    def append(self, item):
        self.model.append([item])
        
    def set_active_from_value(self, value):
        index = 0
        for entry in self.model:
            if entry[0] == value:
                self.cb.set_active(index)
                return
            index += 1
    
    def get_active_text(self):
        return self.cb.get_active_text()
    
    def get_active(self):
        return self.cb.get_active()
    
    def set_active(self, index):
        self.cb.set_active(index)
    
    def set_child_text(self, text):
        self.cb.child.set_text(text)
    
    def set_sensitive(self, val):
        self.cb.set_sensitive(val)

    def set_child_signal(self, signal, callback, userdata=None):
        self.cb.child.connect(signal, callback, userdata)
        
class TextView:
    def __init__(self, widget=None, value="", spellcheck=True):
        if widget is None:
            self.view = gtk.TextView()
        else:
            self.view = widget
        self.buffer = gtk.TextBuffer()
        self.view.set_buffer(self.buffer)
        self.buffer.set_text(value)
        
        if HAS_GTKSPELL and spellcheck:
            try:
                gtkspell.Spell(self.view)
            except Exception, e:
                log.exception(e)
        
    def get_text(self):
        return self.buffer.get_text(
            self.buffer.get_start_iter(), 
            self.buffer.get_end_iter()
        )
        
    def set_text(self, text):
        self.buffer.set_text(text)

    def append_text(self, text):
        self.buffer.set_text(self.get_text() + text)

class SourceView(TextView):
    def __init__(self, widget=None, value=""):
        if HAS_GTKSOURCEVIEW:
            if widget is None:
                self.view = gtksourceview.SourceView(self.buffer)
            else:
                self.view = widget
            self.buffer = gtksourceview.SourceBuffer()
            self.buffer.set_text(value)

            if HAS_GTKSPELL:
                gtkspell.Spell(self.view)

            self.view.show()
        else:
            TextView.__init__(self, widget, value)

class ProgressBar:
    def __init__(self, pbar):
        if pbar is None:
            self.view = gtk.ProgressBar()
        else:
            self.view = pbar
        
        self.timer = None
        
    def start_pulsate(self):
        # Set up an interval to make the progress bar pulse
        # The timeout is removed after the log action finishes
        self.timer = gobject.timeout_add(100, self.update)
    
    def stop_pulsate(self):
        if self.timer:
            gobject.source_remove(self.timer)
        self.timer = None

    def update(self, fraction=None):
        if fraction:
            if self.timer is not None:
                self.stop_pulsate()
                 
            if fraction > 1:
                fraction = 1
            self.view.set_fraction(fraction)
            return False
        else:
            self.view.pulse()
            return True

    def set_text(self, text):
        self.view.set_text(text)

class RevisionSelector:
    """
    Provides a standard way to generate a revision object from the UI.
    
    """

    def __init__(self, container, client, revision=None, 
            url_combobox=None, url_entry=None, url=None, expand=False,
            revision_changed_callback=None):
        """
        @type   container: A gtk container object (i.e. HBox, VBox, Box)
        @param  container: The container that to add this widget
        
        @type   client: VCS client object
        @param  client: A vcs client instance (i.e. rabbitvcs.vcs.VCS())
        
        @type   revision: int
        @param  revision: A revision number to start with
        
        @type   url_combobox: rabbitvcs.ui.widget.ComboBox
        @param  url_combobox: A repository url combobox

        @type   url_entry: gtk.Entry
        @param  url_entry: A repository url entry
        
        @type   url: str
        @param  url: A repository url string
        
        Note: The url fields are required for use with the log browser.  It can
                be excluded.

        """
        
        self.client = client
        self.revision = revision
        self.url_combobox = url_combobox
        self.url_entry = url_entry
        self.url = url
        self.revision_changed_callback = revision_changed_callback
        self.revision_change_inprogress = False
        hbox = gtk.HBox(0, 4)
        
        if self.url_combobox:
            self.url_combobox.cb.connect("changed", self.__on_url_combobox_changed)

        if client.vcs == rabbitvcs.vcs.VCS_GIT:
            self.OPTIONS = [
                _("HEAD"),
                _("Revision"),
                _("Branch")
            ]
        elif client.vcs == rabbitvcs.vcs.VCS_SVN:
            self.OPTIONS = [
                _("HEAD"),
                _("Number")
            ]
            if not client.is_path_repository_url(self.get_url()):
                self.OPTIONS.append(_("Working Copy"))
        
        self.revision_kind_opt = ComboBox(gtk.ComboBox(), self.OPTIONS)
        self.revision_kind_opt.set_active(0)
        self.revision_kind_opt.cb.connect("changed", self.__revision_kind_changed)
        hbox.pack_start(self.revision_kind_opt.cb, False, False, 0)
        
        self.revision_entry = gtk.Entry()
        self.revision_entry.connect("changed", self.__revision_entry_changed)
        hbox.pack_start(self.revision_entry, expand, expand, 0)
        
        self.branch_selector = None
        if client.vcs == rabbitvcs.vcs.VCS_GIT:
            self.branch_selector = GitBranchSelector(hbox, client, self.__branch_selector_changed)
            self.branch_selector.hide()
        
        
        self.revision_browse = gtk.Button()
        revision_browse_image = gtk.Image()
        revision_browse_image.set_from_stock(gtk.STOCK_FIND, 1)
        revision_browse_image.show()
        self.revision_browse.add(revision_browse_image)
        self.revision_browse.connect("clicked", self.__revision_browse_clicked)
        hbox.pack_start(self.revision_browse, False, False, 0)

        if self.revision is not None:
            self.set_kind_number(revision)
        else:
            self.set_kind_head()
        
        self.revision_kind_opt.cb.show()
        self.revision_entry.show()
        self.revision_browse.show()
        hbox.show()
        
        container.add(hbox)
    
    def __revision_browse_clicked(self, widget):
        from rabbitvcs.ui.log import SVNLogDialog, GitLogDialog
        if self.client.vcs == rabbitvcs.vcs.VCS_GIT:
            GitLogDialog(self.get_url(), ok_callback=self.__log_closed)
        elif self.client.vcs == rabbitvcs.vcs.VCS_SVN:
            SVNLogDialog(self.get_url(), ok_callback=self.__log_closed)
    
    def __log_closed(self, data):
        if data is not None:
            self.revision_kind_opt.set_active(1)
            self.revision_entry.set_text(data)

    def __revision_kind_changed(self, widget):
        self.determine_widget_sensitivity()
        
        if self.revision_changed_callback:
            self.revision_change_inprogress = True
            gobject.timeout_add(400, self.__revision_changed_callback, self)

    def __revision_entry_changed(self, widget):
        if self.revision_changed_callback and not self.revision_change_inprogress:
            self.revision_change_inprogress = True
            gobject.timeout_add(400, self.__revision_changed_callback, self)

    def __revision_changed_callback(self, *args, **kwargs):
        self.revision_change_inprogress = False
        self.revision_changed_callback(*args, **kwargs)

    def __on_url_combobox_changed(self, widget):
        self.determine_widget_sensitivity()

    def determine_widget_sensitivity(self):
        index = self.revision_kind_opt.get_active()
        
        allow_revision_browse = True

        # Default to showing the revision entry
        self.hide_branch_selector()
        
        # Only allow number entry if "Number" is selected
        if index == 1:
            self.revision_entry.set_sensitive(True)
        elif index == 2:
            if self.client.vcs == rabbitvcs.vcs.VCS_GIT:
                self.revision_entry.set_sensitive(True)
                allow_revision_browse = False

                # If showing the "Branch" option, show the branch selector
                self.show_branch_selector()
        else:
            self.revision_entry.set_text("")
            self.revision_entry.set_sensitive(False)

        # Only allow browsing if a URL is provided
        if self.get_url() and allow_revision_browse:
            self.revision_browse.set_sensitive(True)
        else:
            self.revision_browse.set_sensitive(False)
    
    def get_url(self):
        if self.url_combobox:
            return self.url_combobox.get_active_text()
        elif self.url_entry:
            return self.url_entry.get_text()
        elif self.url:
            return self.url
        else:
            return ""

    def set_url(self, url):
        self.url = url

    def get_revision_object(self):
        """
        @rtype  rabbitvcs.vcs.###.Revision
        @return A rabbitvcs revision object
        
        index=0     HEAD
        index=1     Revision Number
        index=2
            SVN     Working Copy
            Git     Branch Selector
        
        """
        index = self.revision_kind_opt.get_active()
        
        if index == 0:
            return self.client.revision("head")
        elif index == 1:
            if self.client.vcs == rabbitvcs.vcs.VCS_SVN:
                return self.client.revision("number", self.revision_entry.get_text())
            elif self.client.vcs == rabbitvcs.vcs.VCS_GIT:
                return self.client.revision(self.revision_entry.get_text())
        elif index == 2:
            if self.client.vcs == rabbitvcs.vcs.VCS_SVN:
                return self.client.revision("working")
            elif self.client.vcs == rabbitvcs.vcs.VCS_GIT:
                return self.client.revision(self.branch_selector.get_branch())

    def set_kind_head(self):
        self.revision_kind_opt.set_active(0)
        self.determine_widget_sensitivity()

    def set_kind_number(self, number=None):
        self.revision_kind_opt.set_active(1)
        if number is not None:
            self.revision_entry.set_text(str(number))
        self.determine_widget_sensitivity()

    def set_kind_working(self):
        self.revision_kind_opt.set_active(2)
        self.determine_widget_sensitivity()

    def show_branch_selector(self):
        if self.branch_selector:
            self.revision_entry.hide()
            self.revision_browse.hide()
            self.branch_selector.show()

    def hide_branch_selector(self):
        if self.branch_selector:
            self.branch_selector.hide()
            self.revision_entry.show()
            self.revision_browse.show()

    def __branch_selector_changed(self, branch):
        self.__revision_entry_changed(self.revision_entry)

class KeyValueTable(gtk.Table):
    """
    Simple extension of a GTK table to display a two-column table of information
    with labels.
    """
    
    default_col_spacing = 12
    default_row_spacing = 6
        
    def __init__(self, stuff):
        """
        @param stuff: a list of two-element tuples - the first element of the
                      tuple is the key/label, and the second element is the
                      information
        """
        if not stuff or len(stuff) == 0:
            super(KeyValueTable, self).__init__()
        else:
            super(KeyValueTable, self).__init__(len(stuff), 2)
            
            row = 0
            
            for key, value in stuff:
                label_key = gtk.Label("<b>%s:</b>" % key)
                label_key.set_properties(xalign=0, use_markup=True)
                
                label_value = gtk.Label("%s" % value)
                if HAS_PANGO:
                    label_value.set_properties(xalign=0,                    \
                                           ellipsize=PANGO_ELLIPSIZE_MIDDLE, \
                                           selectable=True)
                else:
                    label_value.set_properties(xalign=0,selectable=True)
                    
                self.attach(label_key,
                             0,1,
                             row, row+1,
                             xoptions=GTK_FILL)
    
                self.attach(label_value,
                             1,2,
                             row, row+1,
                             xoptions=GTK_FILL|GTK_EXPAND)
                        
                label_key.show()
                label_value.show()
                
                row += 1
        
        self.set_col_spacings(self.default_col_spacing)
        self.set_row_spacings(self.default_row_spacing)

class GitRepositorySelector:
    def __init__(self, container, git, changed_callback=None):
        self.git = git
        self.changed_callback = changed_callback
        
        vbox = gtk.VBox(False, 4)
        
        # Set up the Repository Line
        label = gtk.Label(_("Repository:"))
        label.set_size_request(90, -1)
        label.set_justify(gtk.JUSTIFY_LEFT)

        tmp_repos = []
        for item in self.git.remote_list():
            tmp_repos.append(item["name"])
        self.repository_opt = ComboBox(gtk.ComboBoxEntry(), tmp_repos)
        self.repository_opt.set_active(0)
        self.repository_opt.cb.connect("changed", self.__repository_changed)
        self.repository_opt.cb.set_size_request(175, -1)
        
        hbox = gtk.HBox(False, 0)
        hbox.pack_start(label, False, False, 0)
        hbox.pack_start(self.repository_opt.cb, False, False, 0)
        vbox.pack_start(hbox, False, False, 0)


        # Set up the Branch line
        label = gtk.Label(_("Branch:"))
        label.set_size_request(90, -1)
        label.set_justify(gtk.JUSTIFY_LEFT)

        tmp_branches = []
        active_branch_index = 0
        index = 0
        for item in self.git.branch_list():
            tmp_branches.append(item.name)
            
            if item.tracking:
                active_branch_index = index
            
            index += 1
            
        self.branch_opt = ComboBox(gtk.ComboBoxEntry(), tmp_branches)
        self.branch_opt.set_active(active_branch_index)
        self.branch_opt.cb.connect("changed", self.__branch_changed)
        self.branch_opt.cb.set_size_request(175, -1)
        
        hbox = gtk.HBox(False, 0)
        hbox.pack_start(label, False, False, 0)
        hbox.pack_start(self.branch_opt.cb, False, False, 0)
        vbox.pack_start(hbox, False, False, 0)
        
        # Set up the Host line
        label = gtk.Label(_("Host:"))
        label.set_justify(gtk.JUSTIFY_LEFT)
        label.set_size_request(90, -1)
        self.host = gtk.Label()
        self.host.set_justify(gtk.JUSTIFY_LEFT)
        hbox = gtk.HBox(False, 0)
        hbox.pack_start(label, False, False, 0)
        hbox.pack_start(self.host, False, False, 0)
        vbox.pack_start(hbox, False, False, 4)

        vbox.show_all()
        container.add(vbox)
        
        self.__update_host()
    
    def __update_host(self):
        repo = "origin"
        self.host.set_text(self.git.config.get('remote "%s"' % repo, "url"))
    
    def __repository_changed(self, repository_opt):
        if self.changed_callback:
            self.changed_callback(repository_opt.get_active_text(), self.branch_opt.get_active_text())
   
    def __branch_changed(self, branch_opt):
        if self.changed_callback:
            self.changed_callback(self.repository_opt.get_active_text(), self.branch_opt.get_active_text())

class GitBranchSelector:
    def __init__(self, container, git, changed_callback=None):
        self.git = git
        self.changed_callback = changed_callback
        
        self.vbox = gtk.VBox(False, 4)

        tmp_branches = []
        active = 0
        index = 0
        for item in self.git.branch_list():
            tmp_branches.append(item.name)
            if self.git.is_tracking(item.name):
                active = index
            index += 1

        self.branch_opt = ComboBox(gtk.ComboBoxEntry(), tmp_branches)
        self.branch_opt.set_active(active)
        self.branch_opt.cb.connect("changed", self.__branch_changed)
        self.branch_opt.cb.set_size_request(175, -1)

        hbox = gtk.HBox(False, 0)
        hbox.pack_start(self.branch_opt.cb, False, False, 0)
        self.vbox.pack_start(hbox, False, False, 0)
        
        self.vbox.show_all()
        container.add(self.vbox)

    def append(self, widget):
        self.vbox.pack_start(widget, False, False, 0)

    def get_branch(self):
        return self.branch_opt.get_active_text()

    def __branch_changed(self, branch_opt):
        pass
    
    def show(self):
        self.vbox.show_all()
    
    def hide(self):
        self.vbox.hide()

class MultiFileTextEditor:
    """
    Edit a set of text/config/ignore files
    """
    
    def __init__(self, container, label, combobox_labels, combobox_paths, show_add_line=True, line_content=""):
        self.container = container
        self.label = label
        self.combobox_labels = combobox_labels
        self.combobox_paths = combobox_paths
        
        self.cache = {}
        
        self.last_path = None
                
        self.combobox = ComboBox(gtk.ComboBox(), self.combobox_labels)
        self.combobox.cb.connect("changed", self.__combobox_changed)
        self.combobox.cb.set_size_request(175, -1)
        
        self.textview = TextView(gtk.TextView())

        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        scrolled_window.add(self.textview.view)
        scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled_window.set_size_request(320, 150)
        
        vbox = gtk.VBox(False, 6)
        
        hbox = gtk.HBox(False, 3)
        combo_label = gtk.Label(label)
        combo_label.set_alignment(0, 0.5)
        combo_label.set_size_request(130, -1)
        hbox.pack_start(combo_label, False, False, 0)
        hbox.pack_start(self.combobox.cb, True, True, 0)
        vbox.pack_start(hbox, False, False, 0)
        
        if show_add_line:
            hbox = gtk.HBox(False, 3)
            add_label = gtk.Label(_("Add line:"))
            add_label.set_alignment(0, 0.5)
            add_label.set_size_request(130, -1)
            self.add_entry = gtk.Entry()
            self.add_entry.set_text(line_content)
            add_button = gtk.Button(_("Add"))
            add_button.connect("clicked", self.__add_button_clicked)
            hbox.pack_start(add_label, False, False, 0)
            hbox.pack_start(self.add_entry, True, True, 0)
            hbox.pack_start(add_button, False, False, 0)
            vbox.pack_start(hbox, False, False, 0)
        
        vbox.pack_start(scrolled_window, True, True, 0)
        vbox.show_all()

        self.combobox.set_active(0)

        container.add(vbox)
        
    def __combobox_changed(self, widget):
        index = self.combobox.get_active()
        path = self.combobox_paths[index]

        if not self.last_path:
            self.last_path = path
        
        self.cache[self.last_path] = self.textview.get_text()
        self.last_path = path
        
        self.load_file(path)
    
    def __add_button_clicked(self, widget):
        text = self.add_entry.get_text()
        self.add_line(text)
        self.add_entry.set_text("")
    
    def add_line(self, text):
        current_text = self.textview.get_text()
        if current_text:
            current_text += "\n" + text
        else:
            current_text = text
        
        self.textview.set_text(current_text)
    
    def load_file(self, path):
        if os.path.exists(path):
            fh = open(path, "r")
            self.textview.set_text(fh.read())
            fh.close()
        else:
            self.textview.set_text("")

    def save(self, path=None):
        if path:
            paths = [path]
        else:
            paths = self.combobox_paths

        index = self.combobox.get_active()
        current_path = self.combobox_paths[index]
        self.cache[current_path] = self.textview.get_text()
        
        for tmppath in paths:
            if tmppath in self.cache:
                if not os.path.exists(os.path.dirname(tmppath)):
                    os.mkdir(os.path.dirname(tmppath))

                fh = open(tmppath, "w")
                fh.write(self.cache[tmppath])
                fh.close()



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import pygtk
import gobject
import gtk

from rabbitvcs.ui import InterfaceView
from rabbitvcs.ui.action import SVNAction
import rabbitvcs.ui.widget
import rabbitvcs.ui.dialog
import rabbitvcs.util.helper

from rabbitvcs import gettext
_ = gettext.gettext

class SVNSwitch(InterfaceView):
    def __init__(self, path, revision=None):
        InterfaceView.__init__(self, "switch", "Switch")

        self.path = path
        self.vcs = rabbitvcs.vcs.VCS()
        self.svn = self.vcs.svn()
        
        self.get_widget("path").set_text(self.path)
        self.repositories = rabbitvcs.ui.widget.ComboBox(
            self.get_widget("repositories"), 
            rabbitvcs.util.helper.get_repository_paths()
        )

        self.revision_selector = rabbitvcs.ui.widget.RevisionSelector(
            self.get_widget("revision_container"),
            self.svn,
            revision=revision,
            url_combobox=self.repositories,
            expand=True
        )
        
        self.repositories.set_child_text(rabbitvcs.util.helper.unquote_url(self.svn.get_repo_url(self.path)))

    def on_ok_clicked(self, widget):
        url = self.repositories.get_active_text()
        
        if not url or not self.path:
            rabbitvcs.ui.dialog.MessageBox(_("The repository location is a required field."))
            return

        revision = self.revision_selector.get_revision_object()
        self.hide()
        self.action = rabbitvcs.ui.action.SVNAction(
            self.svn,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        
        self.action.append(self.action.set_header, _("Switch"))
        self.action.append(self.action.set_status, _("Running Switch Command..."))
        self.action.append(rabbitvcs.util.helper.save_repository_path, url)
        self.action.append(
            self.svn.switch,
            self.path,
            rabbitvcs.util.helper.quote_url(url),
            revision=revision
        )
        self.action.append(self.action.set_status, _("Completed Switch"))
        self.action.append(self.action.finish)
        self.action.start()

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNSwitch
}

def switch_factory(path, revision=None):
    guess = rabbitvcs.vcs.guess(path)
    return classes_map[guess["vcs"]](path, revision)

if __name__ == "__main__":
    from rabbitvcs.ui import main, REVISION_OPT
    (options, args) = main(
        [REVISION_OPT],
        usage="Usage: rabbitvcs switch [url]"
    )
            
    window = switch_factory(args[0], revision=options.revision)
    window.register_gtk_quit()
    gtk.main()



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import thread

import pygtk
import gobject
import gtk

from rabbitvcs.ui import InterfaceView, InterfaceNonView
from rabbitvcs.ui.add import Add
from rabbitvcs.ui.action import SVNAction
import rabbitvcs.ui.widget
import rabbitvcs.ui.dialog
import rabbitvcs.ui.action
import rabbitvcs.util.helper
from rabbitvcs.util.log import Log

log = Log("rabbitvcs.ui.unlock")

from rabbitvcs import gettext
_ = gettext.gettext

class SVNUnlock(Add):
    def __init__(self, paths, base_dir):
        InterfaceView.__init__(self, "add", "Add")

        self.window = self.get_widget("Add")
        self.window.set_title(_("Unlock"))

        self.paths = paths
        self.base_dir = base_dir
        self.last_row_clicked = None
        self.vcs = rabbitvcs.vcs.VCS()
        self.svn = self.vcs.svn()
        self.items = None
        self.statuses = None
        self.files_table = rabbitvcs.ui.widget.Table(
            self.get_widget("files_table"), 
            [gobject.TYPE_BOOLEAN, rabbitvcs.ui.widget.TYPE_PATH, 
                gobject.TYPE_STRING], 
            [rabbitvcs.ui.widget.TOGGLE_BUTTON, _("Path"), _("Extension")],
            filters=[{
                "callback": rabbitvcs.ui.widget.path_filter,
                "user_data": {
                    "base_dir": base_dir,
                    "column": 1
                }
            }],
            callbacks={
                "row-activated":  self.on_files_table_row_activated,
                "mouse-event":   self.on_files_table_mouse_event,
                "key-event":     self.on_files_table_key_event
            }
        )

        self.initialize_items()
        
        
    #
    # Helpers
    #

    def load(self):
        gtk.gdk.threads_enter()
        self.get_widget("status").set_text(_("Loading..."))
        self.items = self.vcs.get_items(self.paths, self.statuses)
        self.populate_files_table()
        gtk.gdk.threads_leave()
    
    def populate_files_table(self):
        self.files_table.clear()

        found = 0
        for item in self.items:
            # FIXME: ...
            if item.simple_content_status() in (rabbitvcs.vcs.status.status_unversioned, rabbitvcs.vcs.status.status_ignored):
                continue
                
            if not self.svn.is_locked(item.path):
                continue
        
            self.files_table.append([
                True, 
                item.path, 
                rabbitvcs.util.helper.get_file_extension(item.path)
            ])
            found += 1
            
        self.get_widget("status").set_text(_("Found %d item(s)") % found)
    
    #
    # UI Signal Callbacks
    #
             
    def on_ok_clicked(self, widget):
        items = self.files_table.get_activated_rows(1)
        if not items:
            self.close()
            return
        self.hide()

        self.action = rabbitvcs.ui.action.SVNAction(
            self.svn,
            register_gtk_quit=self.gtk_quit_is_set()
        )
        
        self.action.append(self.action.set_header, _("Unlock"))
        self.action.append(self.action.set_status, _("Running Unlock Command..."))
        for item in items:
            self.action.append(self.svn.unlock, item, force=True)
        self.action.append(self.action.set_status, _("Completed Unlock"))
        self.action.append(self.action.finish)
        self.action.start()

class SVNUnlockQuiet:
    """
    This class provides a handler to unlock functionality.
    
    """

    def __init__(self, paths, base_dir=None):
        self.paths = paths
        self.vcs = rabbitvcs.vcs.VCS()
        self.svn = self.vcs.svn()

        for path in self.paths:
            self.svn.unlock(path, force=True)

classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNUnlock
}

quiet_classes_map = {
    rabbitvcs.vcs.VCS_SVN: SVNUnlockQuiet
}

def unlock_factory(cmap, paths, base_dir=None):
    guess = rabbitvcs.vcs.guess(paths[0])
    return cmap[guess["vcs"]](paths, base_dir)

if __name__ == "__main__":
    from rabbitvcs.ui import main, BASEDIR_OPT, QUIET_OPT
    (options, paths) = main(
        [BASEDIR_OPT, QUIET_OPT],
        usage="Usage: rabbitvcs unlock [path1] [path2] ..."
    )
        
    if options.quiet:
        unlock_factory(quiet_classes_map, paths)
    else:
        window = unlock_factory(classes_map, paths, options.base_dir)
        window.register_gtk_quit()
        gtk.main()



#
# Copyright (C) 2009 Jason Heeris <jason.heeris@gmail.com>
# Copyright (C) 2009 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2009 by Adam Plumb <adamplumb@gmail.com>#
#
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

""" The checker service for RabbitVCS background status checks.

This file can be run as a Python script, in which case it starts a background
VCS status checking service that can be called via DBUS. It also contains class
definitions to call these methods from within a separate Python process.

This currently works like so:

    1. Nautilus loads our extension, RabbitVCS
    2. RabbitVCS creates a StatusCheckerStub
    3. StatusCheckerStub calls start(), which is a wrapper for a more general
       service starter convenience method
    4. The service starter method looks for a DBUS object with the given service
       name and object path; if none is found, it creates it by running this
       script

RabbitVCS can then call the stub methods, getting status info via the
CheckStatus method itself, or more likely from a callback upon completion of a
status check.

NOTE: as a general rule, the data piped between processes or sent over DBUS
should be kept to a minimum. Use convenience methods to condense and summarise
data wherever possible (this is the case in the actual status cache and checker
code).
"""

import os, os.path
import sys
import simplejson

try:
    from gi.repository import GObject as gobject
except ImportError:
    import gobject
    
try:
    from gi.repository import GLib as glib
except:
    import glib

import dbus
import dbus.glib # FIXME: this might actually already set the default loop
import dbus.mainloop.glib
import dbus.service

import rabbitvcs.util.decorators
import rabbitvcs.util._locale
import rabbitvcs.util.helper
import rabbitvcs.services.service
from rabbitvcs.services.statuschecker import StatusChecker

import rabbitvcs.vcs.status

from rabbitvcs.util.log import Log
log = Log("rabbitvcs.services.checkerservice")

from rabbitvcs import version as SERVICE_VERSION

INTERFACE = "org.google.code.rabbitvcs.StatusChecker"
OBJECT_PATH = "/org/google/code/rabbitvcs/StatusChecker"
SERVICE = "org.google.code.rabbitvcs.RabbitVCS.Checker"
TIMEOUT = 60*15*100 # seconds

def find_class(module, name):
    """ Given a module name and a class name, return the actual type object.
    """
    # From Python stdlib pickle module source
    __import__(module)
    mod = sys.modules[module]
    klass = getattr(mod, name)
    return klass

def encode_status(status):
    """ Before encoding a status object to JSON, we need to turn it into
    something simpler.
    """
    return status.__getstate__()

def decode_status(json_dict):
    """ Once we get a JSON encoded string out the other side of DBUS, we need to
    reconstitute the original object. This method is based on the pickle module
    in the Python stdlib.
    """    
    cl = find_class(json_dict['__module__'], json_dict['__type__'])
    st = None
    if cl in rabbitvcs.vcs.status.STATUS_TYPES:
        st = cl.__new__(cl)
        st.__setstate__(json_dict)
    elif json_dict.has_key('path'):
        log.warning("Could not deduce status class: %s" % json_dict['__type__'])
        st = rabbitvcs.vcs.status.Status.status_error(json_dict['path'])
    else:
        raise TypeError("RabbitVCS status object has no path")
    return st

class StatusCheckerService(dbus.service.Object):
    """ StatusCheckerService objects wrap a StatusCheckerPlus instance,
    exporting methods that can be called via DBUS.

    There should only be a single such object running in a separate process from
    the GUI (ie. do not create this in the Nautilus extension code, you should
    use a StatusCheckerStub there instead).
    """

    def __init__(self, connection, mainloop):
        """ Creates a new status checker wrapper service, with the given DBUS
        connection.

        The mainloop argument is needed for process management (eg. calling
        Quit() for graceful exiting).

        @param connection: the DBUS connection (eg. session bus, system bus)
        @type connection: a DBUS connection object

        @param mainloop: the main loop that DBUS is using
        @type mainloop: any main loop with a quit() method
        """
        dbus.service.Object.__init__(self, connection, OBJECT_PATH)
        
        self.encoder = simplejson.JSONEncoder(default=encode_status,
                                              separators=(',', ':'))
        
        self.mainloop = mainloop

        # Start the status checking daemon so we can do requests in the
        # background
        self.status_checker = StatusChecker()

    @dbus.service.method(INTERFACE)
    def ExtraInformation(self):
        return self.status_checker.extra_info()

    @dbus.service.method(INTERFACE)
    def MemoryUsage(self):
        own_mem = rabbitvcs.util.helper.process_memory(os.getpid())
        checker_mem = self.status_checker.get_memory_usage()

        return own_mem + checker_mem

    @dbus.service.method(INTERFACE)
    def PID(self):
        return os.getpid()

    @dbus.service.method(INTERFACE)
    def CheckerType(self):
        return self.status_checker.CHECKER_NAME

    @dbus.service.method(INTERFACE, in_signature='sbbb', out_signature='s')
    def CheckStatus(self, path, recurse=False, invalidate=False,
                      summary=False):
        """ Requests a status check from the underlying status checker.
        """
        status = self.status_checker.check_status(unicode(path),
                                                  recurse=recurse,
                                                  summary=summary,
                                                  invalidate=invalidate)
        
        return self.encoder.encode(status)

    @dbus.service.method(INTERFACE, in_signature='as', out_signature='s')
    def GenerateMenuConditions(self, paths):
        upaths = []
        for path in paths:
            upaths.append(unicode(path))
    
        path_dict = self.status_checker.generate_menu_conditions(upaths)
        return simplejson.dumps(path_dict)

    @dbus.service.method(INTERFACE)
    def CheckVersionOrDie(self, version):
        """
        If the version passed does not match the version of RabbitVCS available
        when this service started, the service will exit. The return value is
        None if the versions match, else it's the PID of the service (useful for
        waiting for the process to exit).
        """
        if not self.CheckVersion(version):
            log.warning("Version mismatch, quitting checker service " \
                        "(service: %s, extension: %s)" \
                        % (SERVICE_VERSION, version))
            return self.Quit()

        return None

    @dbus.service.method(INTERFACE)
    def CheckVersion(self, version):
        """
        Return True iff the version of RabbitVCS imported by this service is the
        same as that passed in (ie. used by extension code).
        """
        return version == SERVICE_VERSION

    @dbus.service.method(INTERFACE)
    def Quit(self):
        """ Quits the service, performing any necessary cleanup operations.

        You can call this from the command line with:

        dbus-send --print-reply \
        --dest=org.google.code.rabbitvcs.RabbitVCS.Checker \
        /org/google/code/rabbitvcs/StatusChecker \
        org.google.code.rabbitvcs.StatusChecker.Quit

        If calling this programmatically, then you can do "os.waitpid(pid, 0)"
        on the returned PID to prevent a zombie process.
        """
        self.status_checker.quit()
        log.debug("Quitting main loop...")
        self.mainloop.quit()
        return self.PID()


class StatusCheckerStub:
    """ StatusCheckerStub objects contain methods that call an actual status
    checker running in another process.

    These objects should be created by the GUI as needed (eg. the nautilus
    extension code).

    The inter-process communication is via DBUS.
    """

    def __init__(self):
        """ Creates an object that can call the VCS status checker via DBUS.

        If there is not already a DBUS object with the path "OBJECT_PATH", we
        create one by starting a new Python process that runs this file.
        """
        # We need this to for the client to be able to do asynchronous calls
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self.session_bus = dbus.SessionBus()
        self.decoder = simplejson.JSONDecoder(object_hook=decode_status)
        self.status_checker = None
        start()
        self._connect_to_checker()

    def _connect_to_checker(self):

        # Start the status checker, if it's not running this should start it up.
        # Otherwise it leaves it alone.
        # start()

        # Try to get a new checker
        try:
            self.status_checker = self.session_bus.get_object(SERVICE,
                                                              OBJECT_PATH)
        except dbus.DBusException, ex:
            # There is not much we should do about this...
            log.exception(ex)

    def assert_version(self, version):
        """
        This will use the CheckVersionOrDie method to ensure that either the
        checker service currently running has the correct version, or that it
        is quit and restarted.
        
        Note that if the version of the newly started checker still doesn't
        match, nothing is done. 
        """
        try:
            pid = self.status_checker.CheckVersionOrDie(version)
        except dbus.DBusException, ex:
            log.exception(ex)
            self._connect_to_checker()
        else:
            if pid is not None:
                try:
                    os.waitpid(pid, 0)
                except OSError:
                    # Process already gone...
                    pass
                start()
                self._connect_to_checker()
                
                try:
                    if not self.status_checker.CheckVersion(version):
                        log.warning("Version mismatch even after restart!")
                except dbus.DBusException, ex:
                    log.exception(ex)
                    self._connect_to_checker()
                    

    def check_status_now(self, path, recurse=False, invalidate=False,
                       summary=False):
        
        status = None
                
        try:
            json_status = self.status_checker.CheckStatus(path,
                                                          recurse, invalidate,
                                                          summary,
                                                          dbus_interface=INTERFACE,
                                                          timeout=TIMEOUT)
            status = self.decoder.decode(json_status)
            # Test client error problems :)
            # raise dbus.DBusException("Test")
        except dbus.DBusException, ex:
            log.exception(ex)

            status = rabbitvcs.vcs.status.Status.status_error(path)

            # Try to reconnect
            self._connect_to_checker()

        return status
        
    def check_status_later(self, path, callback, recurse=False,
                           invalidate=False, summary=False):
        
        def real_reply_handler(json_status):
            # Note that this a closure referring to the outer functions callback
            # parameter
            status = self.decoder.decode(json_status)
            assert status.path == path, "Status check returned the wrong path "\
                                        "(asked about %s, got back %s)" % \
                                        (path, status.path)
            callback(status)
        
        def reply_handler(*args, **kwargs):
            # The callback should be performed as a low priority task, so we
            # keep Nautilus as responsive as possible.
            gobject.idle_add(real_reply_handler, *args, **kwargs)
        
        def error_handler(dbus_ex):
            log.exception(dbus_ex)
            self._connect_to_checker()
            callback(rabbitvcs.vcs.status.Status.status_error(path))
        
        try:
            self.status_checker.CheckStatus(path,
                                            recurse, invalidate,
                                            summary,
                                            dbus_interface=INTERFACE,
                                            timeout=TIMEOUT,
                                            reply_handler=reply_handler,
                                            error_handler=error_handler)
        except dbus.DBusException, ex:
            log.exception(ex)
            callback(rabbitvcs.vcs.status.Status.status_error(path))
            # Try to reconnect
            self._connect_to_checker()

    # @rabbitvcs.util.decorators.deprecated
    # Can't decide whether this should be deprecated or not... -JH
    def check_status(self, path, recurse=False, invalidate=False,
                       summary=False, callback=None):
        """ Check the VCS status of the given path.

        This is a pass-through method to the check_status method of the DBUS
        service (which is, in turn, a wrapper around the real status checker).
        """
        if callback:
            gobject.idle_add(self.check_status_later,
                     path, callback, recurse, invalidate, summary)
            return rabbitvcs.vcs.status.Status.status_calc(path)
        else:
            return self.check_status_now(path, recurse, invalidate, summary)

    def generate_menu_conditions(self, provider, base_dir, paths, callback):
    
        def real_reply_handler(json):
            # Note that this a closure referring to the outer functions callback
            # parameter
            path_dict = simplejson.loads(json)
            callback(provider, base_dir, paths, path_dict)
        
        def reply_handler(*args, **kwargs):
            # The callback should be performed as a low priority task, so we
            # keep Nautilus as responsive as possible.
            gobject.idle_add(real_reply_handler, *args, **kwargs)
        
        def error_handler(dbus_ex):
            log.exception(dbus_ex)
            self._connect_to_checker()
            callback(provider, base_dir, paths, {})
        
        try:
            self.status_checker.GenerateMenuConditions(paths,
                                            dbus_interface=INTERFACE,
                                            timeout=TIMEOUT,
                                            reply_handler=reply_handler,
                                            error_handler=error_handler)
        except dbus.DBusException, ex:
            log.exception(ex)
            callback(provider, base_dir, paths, {})
            # Try to reconnect
            self._connect_to_checker()

    def generate_menu_conditions_async(self, provider, base_dir, paths, callback):
        gobject.idle_add(self.generate_menu_conditions, provider, base_dir, paths, callback)
        return {}

def start():
    """ Starts the checker service, via the utility method in "service.py". """
    rabbitvcs.services.service.start_service(os.path.abspath(__file__), SERVICE,
                                             OBJECT_PATH)

def Main():
    """ The main point of entry for the checker service.

    This will set up the DBUS and glib extensions, the gobject/glib main loop,
    and start the service.
    """
    global log
    log = Log("rabbitvcs.services.checkerservice:main")
    log.debug("Checker: starting service: %s (%s)" % (OBJECT_PATH, os.getpid()))

    # We need this to for the client to be able to do asynchronous calls
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

    # The following calls are required to make DBus thread-aware and therefore
    # support the ability run threads.
    gobject.threads_init()
    dbus.glib.threads_init()

    # This registers our service name with the bus
    session_bus = dbus.SessionBus()
    service_name = dbus.service.BusName(SERVICE, session_bus)

    mainloop = gobject.MainLoop()

    checker_service = StatusCheckerService(session_bus, mainloop)

    gobject.idle_add(sys.stdout.write, "Started status checker service\n")
    gobject.idle_add(sys.stdout.flush)

    mainloop.run()

    log.debug("Checker: ended service: %s (%s)" % (OBJECT_PATH, os.getpid()))

if __name__ == "__main__":
    rabbitvcs.util._locale.initialize_locale()

    # import cProfile
    # import rabbitvcs.util.helper
    # profile_data_file = os.path.join(
    #                        rabbitvcs.util.helper.get_home_folder(),
    #                        "checkerservice.stats")
    # cProfile.run("Main()", profile_data_file)

    Main()



#
# This is an extension to the Nautilus file manager to allow better 
# integration with the Subversion source control system.
# 
# Copyright (C) 2006-2008 by Jason Field <jason@jasonfield.com>
# Copyright (C) 2007-2008 by Bruce van der Kooij <brucevdkooij@gmail.com>
# Copyright (C) 2008-2010 by Adam Plumb <adamplumb@gmail.com>
# 
# RabbitVCS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RabbitVCS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RabbitVCS;  If not, see <http://www.gnu.org/licenses/>.
#

import os.path
from rabbitvcs import gettext
_ = gettext.gettext

from rabbitvcs.util.log import Log
logger = Log("rabbitvcs.vcs")

from rabbitvcs.util.helper import get_exclude_paths

EXT_UTIL_ERROR = _("The output from '%s' was not able to be processed.\n%s")

VCS_SVN = 'svn'
VCS_GIT = 'git'
VCS_MERCURIAL = 'mercurial'
VCS_DUMMY = 'unknown'

def guess(path):
    # Determine the VCS instance based on the path
    if path:
        path_to_check = path
        folders = {
            ".svn": VCS_SVN,
            ".git": VCS_GIT,
            ".hg": VCS_MERCURIAL,
            ".bzr": VCS_DUMMY,
            ".CVS": VCS_DUMMY
        }
        
        while path_to_check != "/" and path_to_check != "":
            for folder, client in folders.items():
                if os.path.isdir(os.path.join(path_to_check, folder)):
                    cache = {
                        "vcs": client,
                        "repo_path": path_to_check
                    }
                    return cache                
            path_to_check = os.path.split(path_to_check)[0]

    # Attempt 2 - assume it's a path like "local.txt@1"
    path_to_check = "./" + path.split("@")[0]

    while path_to_check != "/" and path_to_check != "":
        for folder, client in folders.items():
            if os.path.isdir(os.path.join(path_to_check, folder)):
                cache = {
                    "vcs": client,
                    "repo_path": path_to_check
                }
                return cache                
        path_to_check = os.path.split(path_to_check)[0]            


    return {
        "vcs": VCS_DUMMY,
        "repo_path": path
    }

class VCS:
    clients = {}
    exclude_paths = []
    
    def __init__(self):
        self.exclude_paths = get_exclude_paths()
    
    def dummy(self):
        if VCS_DUMMY in self.clients:
            return self.clients[VCS_DUMMY]
        else:
            from rabbitvcs.vcs.dummy import Dummy
            self.clients[VCS_DUMMY] = Dummy()
            return self.clients[VCS_DUMMY]
    
    def svn(self):
        if VCS_SVN in self.clients:
            return self.clients[VCS_SVN]
        else:
            try:
                from rabbitvcs.vcs.svn import SVN
                self.clients[VCS_SVN] = SVN()
                return self.clients[VCS_SVN]
            except Exception, e:
                logger.debug("Unable to load SVN module: %s" % e)
                logger.exception(e)
                self.clients[VCS_SVN] = self.dummy()
                return self.clients[VCS_SVN]

    def git(self, path=None, is_repo_path=False):
        if VCS_GIT in self.clients:
            git = self.clients[VCS_GIT]

            if path:
                if is_repo_path:
                    git.set_repository(path)
                else:
                    repo_path = git.find_repository_path(path)
                    git.set_repository(repo_path)

            return git
        else:
            try:
                from rabbitvcs.vcs.git import Git
                git = Git()

                if path:
                    if is_repo_path:
                        git.set_repository(path)
                    else:
                        repo_path = git.find_repository_path(path)
                        git.set_repository(repo_path)
                
                self.clients[VCS_GIT] = git
                return self.clients[VCS_GIT]
            except Exception, e:
                logger.debug("Unable to load Git module: %s" % e)
                logger.exception(e)
                self.clients[VCS_GIT] = self.dummy()
                return self.clients[VCS_GIT]

    def mercurial(self, path=None, is_repo_path=False):
        if VCS_MERCURIAL in self.clients:
            mercurial = self.clients[VCS_MERCURIAL]

            if path:
                if is_repo_path:
                    mercurial.set_repository(path)
                else:
                    repo_path = mercurial.find_repository_path(path)
                    mercurial.set_repository(repo_path)

            return mercurial
        else:
            try:
                from rabbitvcs.vcs.mercurial import Mercurial
                mercurial = Mercurial()

                if path:
                    if is_repo_path:
                        mercurial.set_repository(path)
                    else:
                        repo_path = mercurial.find_repository_path(path)
                        mercurial.set_repository(repo_path)

                self.clients[VCS_MERCURIAL] = mercurial
                return self.clients[VCS_MERCURIAL]
            except Exception, e:
                logger.debug("Unable to load Mercurial module: %s" % e)
                logger.exception(e)
                self.clients[VCS_MERCURIAL] = self.dummy()
                return self.clients[VCS_MERCURIAL]

    def client(self, path, vcs=None):
        if self.should_exclude(path):
            logger.debug("Excluding path: %s" % path)
            return self.dummy()

        # Determine the VCS instance based on the vcs parameter
        if vcs:
            if vcs == VCS_SVN:
                return self.svn()
            elif vcs == VCS_GIT:
                return self.git(path)
            elif vcs == VCS_MERCURIAL:
                return self.mercurial(path)

        guess = self.guess(path)
        if guess["vcs"] == VCS_GIT:
            return self.git(guess["repo_path"], is_repo_path=False)
        elif guess["vcs"] == VCS_SVN:
            return self.svn()
        elif guess["vcs"] == VCS_MERCURIAL:
            return self.mercurial(guess["repo_path"], is_repo_path=False)
        else:
            return self.dummy()
    
    def should_exclude(self, path):
        for exclude_path in self.exclude_paths:
            if path.startswith(exclude_path):
                return True
        
        return False
    
    def guess(self, path):
        return guess(path)
    
    # Methods that call client methods

    def statuses(self, path, recurse=True, invalidate=False):
        client = self.client(path)
        return client.statuses(path, recurse=recurse, invalidate=invalidate)
    
    def status(self, path, summarize=True, invalidate=False):
        client = self.client(path)
        return client.status(path, summarize, invalidate)

    def is_working_copy(self, path):
        client = self.client(path)
        return client.is_working_copy(path)

    def is_in_a_or_a_working_copy(self, path):
        client = self.client(path)
        ret = client.is_in_a_or_a_working_copy(path)
        return ret

    def is_versioned(self, path):
        client = self.client(path)
        return client.is_versioned(path)
    
    def is_locked(self, path):
        client = self.client(path)
        return client.is_locked(path)

    def get_items(self, paths, statuses=[]):
        client = self.client(paths[0])
        return client.get_items(paths, statuses)

    def statuses_for_add(self,paths):
        client = self.client(paths[0])
        return client.STATUSES_FOR_ADD

    def statuses_for_commit(self, paths):
        client = self.client(paths[0])
        return client.STATUSES_FOR_COMMIT

    def statuses_for_revert(self, paths):
        client = self.client(paths[0])
        return client.STATUSES_FOR_REVERT

def create_vcs_instance(path=None, vcs=None):
    """
    Create a VCS instance based on the working copy path
    """
    return VCS()

def guess_vcs(path):
    vcs = VCS()
    return vcs.guess(path)

class ExternalUtilError(Exception):
    """ Represents an error caused by unexpected output from an external
    program.
    """ 
        
    def __init__(self, program, output):
        """ Initialises the error with the external tool and the unexpected
        output.
        """
        Exception.__init__(self,
                           EXT_UTIL_ERROR % (program, output))
        self.program = program
        self.output = output
